# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2014 Wikimedia Foundation Inc.
# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import configparser
import multiprocessing
import os
import re
import time
from unittest import mock

from tests.base import (
    AnsibleZuulTestCase,
    BaseTestCase,
    FIXTURE_DIR,
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)

from zuul.executor.sensors.startingbuilds import StartingBuildsSensor
from zuul.executor.sensors.process import ProcessSensor
from zuul.executor.sensors.ram import RAMSensor
from zuul.executor.server import squash_variables
from zuul.model import NodeSet, Group


class ExecutorReposMixin:
    def assertRepoState(self, repo, state, project, build, number):
        if 'branch' in state:
            self.assertFalse(repo.head.is_detached,
                             'Project %s commit for build %s #%s should '
                             'not have a detached HEAD' % (
                                 project, build, number))
            self.assertEqual(repo.active_branch.name,
                             state['branch'],
                             'Project %s commit for build %s #%s should '
                             'be on the correct branch' % (
                                 project, build, number))
            # Remote 'origin' needs to be kept intact with a bogus URL
            self.assertEqual('file:///dev/null', repo.remotes.origin.url)
            self.assertIn(state['branch'], repo.remotes.origin.refs)
        if 'commit' in state:
            self.assertEqual(state['commit'],
                             str(repo.commit('HEAD')),
                             'Project %s commit for build %s #%s should '
                             'be correct' % (
                                 project, build, number))
        ref = repo.commit('HEAD')
        repo_messages = set(
            [c.message.strip() for c in repo.iter_commits(ref)])
        if 'present' in state:
            for change in state['present']:
                msg = '%s-1' % change.subject
                self.assertTrue(msg in repo_messages,
                                'Project %s for build %s #%s should '
                                'have change %s' % (
                                    project, build, number, change.subject))
        if 'absent' in state:
            for change in state['absent']:
                msg = '%s-1' % change.subject
                self.assertTrue(msg not in repo_messages,
                                'Project %s for build %s #%s should '
                                'not have change %s' % (
                                    project, build, number, change.subject))

    def assertBuildStates(self, states, projects):
        for number, build in enumerate(self.builds):
            work = build.getWorkspaceRepos(projects)
            state = states[number]

            for project in projects:
                self.assertRepoState(work[project], state[project],
                                     project, build, number)

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()


class TestExecutorRepos(ZuulTestCase, ExecutorReposMixin):
    tenant_config_file = 'config/single-tenant/main.yaml'

    log = logging.getLogger("zuul.test.executor")

    @simple_layout('layouts/repo-checkout-two-project.yaml')
    def test_one_branch(self):
        self.executor_server.hold_jobs_in_build = True

        p1 = 'review.example.com/org/project1'
        p2 = 'review.example.com/org/project2'
        projects = [p1, p2]
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(2, len(self.builds), "Two builds are running")

        upstream = self.getUpstreamRepos(projects)
        states = [
            {p1: dict(present=[A], absent=[B], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      branch='master'),
             },
            {p1: dict(present=[A], absent=[B], branch='master'),
             p2: dict(present=[B], absent=[A], branch='master'),
             },
        ]

        self.assertBuildStates(states, projects)

    @simple_layout('layouts/repo-checkout-four-project.yaml')
    def test_multi_branch(self):
        self.executor_server.hold_jobs_in_build = True

        p1 = 'review.example.com/org/project1'
        p2 = 'review.example.com/org/project2'
        p3 = 'review.example.com/org/project3'
        p4 = 'review.example.com/org/project4'
        projects = [p1, p2, p3, p4]

        self.create_branch('org/project2', 'stable/havana')
        self.create_branch('org/project4', 'stable/havana')
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project2', 'stable/havana',
                                           'B')
        C = self.fake_gerrit.addFakeChange('org/project3', 'master', 'C')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(3, len(self.builds), "Three builds are running")

        upstream = self.getUpstreamRepos(projects)
        states = [
            {p1: dict(present=[A], absent=[B, C], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      branch='master'),
             p3: dict(commit=str(upstream[p3].commit('master')),
                      branch='master'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             },
            {p1: dict(present=[A], absent=[B, C], branch='master'),
             p2: dict(present=[B], absent=[A, C], branch='stable/havana'),
             p3: dict(commit=str(upstream[p3].commit('master')),
                      branch='master'),
             p4: dict(commit=str(upstream[p4].commit('stable/havana')),
                      branch='stable/havana'),
             },
            {p1: dict(present=[A], absent=[B, C], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      branch='master'),
             p3: dict(present=[C], absent=[A, B], branch='master'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             },
        ]

        self.assertBuildStates(states, projects)

    @simple_layout('layouts/repo-checkout-six-project.yaml')
    def test_project_override(self):
        self.executor_server.hold_jobs_in_build = True

        p1 = 'review.example.com/org/project1'
        p2 = 'review.example.com/org/project2'
        p3 = 'review.example.com/org/project3'
        p4 = 'review.example.com/org/project4'
        p5 = 'review.example.com/org/project5'
        p6 = 'review.example.com/org/project6'
        projects = [p1, p2, p3, p4, p5, p6]

        self.create_branch('org/project3', 'stable/havana')
        self.create_branch('org/project4', 'stable/havana')
        self.create_branch('org/project6', 'stable/havana')
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project2', 'master', 'C')
        D = self.fake_gerrit.addFakeChange('org/project3', 'stable/havana',
                                           'D')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)
        D.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.fake_gerrit.addEvent(D.addApproval('Approved', 1))
        self.waitUntilSettled()

        self.assertEqual(4, len(self.builds), "Four builds are running")

        upstream = self.getUpstreamRepos(projects)
        states = [
            {p1: dict(present=[A], absent=[B, C, D], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      branch='master'),
             p3: dict(commit=str(upstream[p3].commit('master')),
                      branch='master'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             p5: dict(commit=str(upstream[p5].commit('master')),
                      branch='master'),
             p6: dict(commit=str(upstream[p6].commit('master')),
                      branch='master'),
             },
            {p1: dict(present=[A, B], absent=[C, D], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      branch='master'),
             p3: dict(commit=str(upstream[p3].commit('master')),
                      branch='master'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             p5: dict(commit=str(upstream[p5].commit('master')),
                      branch='master'),
             p6: dict(commit=str(upstream[p6].commit('master')),
                      branch='master'),
             },
            {p1: dict(present=[A, B], absent=[C, D], branch='master'),
             p2: dict(present=[C], absent=[A, B, D], branch='master'),
             p3: dict(commit=str(upstream[p3].commit('master')),
                      branch='master'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             p5: dict(commit=str(upstream[p5].commit('master')),
                      branch='master'),
             p6: dict(commit=str(upstream[p6].commit('master')),
                      branch='master'),
             },
            {p1: dict(present=[A, B], absent=[C, D], branch='master'),
             p2: dict(present=[C], absent=[A, B, D], branch='master'),
             p3: dict(present=[D], absent=[A, B, C],
                      branch='stable/havana'),
             p4: dict(commit=str(upstream[p4].commit('master')),
                      branch='master'),
             p5: dict(commit=str(upstream[p5].commit('master')),
                      branch='master'),
             p6: dict(commit=str(upstream[p6].commit('stable/havana')),
                      branch='stable/havana'),
             },
        ]

        self.assertBuildStates(states, projects)

    def test_periodic_override(self):
        # This test can not use simple_layout because it must start
        # with a configuration which does not include a
        # timer-triggered job so that we have an opportunity to set
        # the hold flag before the first job.

        # This tests that we can override the branch in a timer
        # trigger (mostly to ensure backwards compatability for jobs).
        self.executor_server.hold_jobs_in_build = True

        p1 = 'review.example.com/org/project1'
        projects = [p1]
        self.create_branch('org/project1', 'stable/havana')

        # Start timer trigger - also org/project
        self.commitConfigUpdate('common-config',
                                'layouts/repo-checkout-timer-override.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        # The pipeline triggers every second, so we should have seen
        # several by now.
        time.sleep(5)
        self.waitUntilSettled()

        # Stop queuing timer triggered jobs so that the assertions
        # below don't race against more jobs being queued.
        self.commitConfigUpdate('common-config',
                                'layouts/repo-checkout-no-timer-override.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()
        # If APScheduler is in mid-event when we remove the job, we
        # can end up with one more event firing, so give it an extra
        # second to settle.
        time.sleep(1)
        self.waitUntilSettled()

        self.assertEqual(1, len(self.builds), "One build is running")

        upstream = self.getUpstreamRepos(projects)
        states = [
            {p1: dict(commit=str(upstream[p1].commit('stable/havana')),
                      branch='stable/havana'),
             },
        ]

        self.assertBuildStates(states, projects)

    def test_periodic(self):
        # This test can not use simple_layout because it must start
        # with a configuration which does not include a
        # timer-triggered job so that we have an opportunity to set
        # the hold flag before the first job.
        self.executor_server.hold_jobs_in_build = True
        # Start timer trigger - also org/project
        self.commitConfigUpdate('common-config',
                                'layouts/repo-checkout-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))

        p1 = 'review.example.com/org/project1'
        projects = [p1]
        self.create_branch('org/project1', 'stable/havana')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project1', 'stable/havana'))
        self.waitUntilSettled()

        # The pipeline triggers every second, so we should have seen
        # several by now.
        time.sleep(5)
        self.waitUntilSettled()

        # Stop queuing timer triggered jobs so that the assertions
        # below don't race against more jobs being queued.
        self.commitConfigUpdate('common-config',
                                'layouts/repo-checkout-no-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()
        # If APScheduler is in mid-event when we remove the job, we
        # can end up with one more event firing, so give it an extra
        # second to settle.
        time.sleep(1)
        self.waitUntilSettled()

        self.assertEqual(2, len(self.builds), "Two builds are running")

        upstream = self.getUpstreamRepos(projects)
        states = [
            {p1: dict(commit=str(upstream[p1].commit('stable/havana')),
                      branch='stable/havana'),
             },
            {p1: dict(commit=str(upstream[p1].commit('master')),
                      branch='master'),
             },
        ]
        if self.builds[0].parameters['zuul']['ref'] == 'refs/heads/master':
            states = list(reversed(states))

        self.assertBuildStates(states, projects)

    @simple_layout('layouts/repo-checkout-post.yaml')
    def test_post_and_master_checkout(self):
        self.executor_server.hold_jobs_in_build = True
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p1, p2]
        upstream = self.getUpstreamRepos(projects)

        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        event = A.getRefUpdatedEvent()
        A.setMerged()
        A_commit = str(upstream[p1].commit('master'))
        self.log.debug("A commit: %s" % A_commit)

        # Add another commit to the repo that merged right after this
        # one to make sure that our post job runs with the one that we
        # intended rather than simply the current repo state.
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B',
                                           parent='refs/changes/01/1/1')
        B.setMerged()
        B_commit = str(upstream[p1].commit('master'))
        self.log.debug("B commit: %s" % B_commit)

        self.fake_gerrit.addEvent(event)
        self.waitUntilSettled()

        states = [
            {p1: dict(commit=A_commit,
                      present=[A], absent=[B], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      absent=[A, B], branch='master'),
             },
        ]

        self.assertBuildStates(states, projects)

    @simple_layout('layouts/repo-checkout-tag.yaml')
    def test_tag_checkout(self):
        self.executor_server.hold_jobs_in_build = True
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p1, p2]
        upstream = self.getUpstreamRepos(projects)

        self.create_branch('org/project2', 'stable/havana')
        files = {'README': 'tagged readme'}
        self.addCommitToRepo('org/project2', 'tagged commit',
                             files, branch='stable/havana', tag='test-tag')

        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        states = [
            {p1: dict(present=[A], branch='master'),
             p2: dict(commit=str(upstream[p2].commit('test-tag')),
                      absent=[A]),
             },
        ]

        self.assertBuildStates(states, projects)

    @simple_layout('layouts/reproducible-repo.yaml')
    def test_reproducible_repo(self):
        # Verify that we get exactly the same commit shas when merging
        # items.
        self.executor_server.hold_jobs_in_build = True
        self.hold_jobs_in_queue = True

        project = 'review.example.com/org/project'
        projects = [project]

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        time.sleep(1)
        self.waitUntilSettled()

        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        time.sleep(1)
        self.waitUntilSettled()

        # Four builds running:
        #   job1, job2 for change A
        #   job1, job2 for change A+B
        queue = list(self.executor_api.queued())
        self.assertEqual(len(self.builds), 0)
        self.assertEqual(len(queue), 4)

        for x in range(4):
            self.executor_api.release(queue[x])
            time.sleep(1)
            self.waitUntilSettled()

        self.assertEqual(len(self.builds), 4)
        work = [build.getWorkspaceRepos(projects) for build in self.builds]
        shas = [
            [c.hexsha for c in repos[project].iter_commits(
                repos[project].commit('HEAD'))]
            for repos in work
        ]
        self.log.debug("Shas: %s", shas)

        # Verify that the commit shas are the same for each pair of builds
        self.assertEqual(shas[0], shas[1])
        self.assertEqual(shas[2], shas[3])
        # Verify that every sha in the first 2 builds are in the last
        # two as well
        for sha in shas[0]:
            self.assertTrue(sha in shas[2])

        states = [
            {project: dict(present=[A], absent=[B], branch='master')},
            {project: dict(present=[A], absent=[B], branch='master')},
            {project: dict(present=[A, B], absent=[], branch='master')},
            {project: dict(present=[A, B], absent=[], branch='master')},
        ]
        self.assertBuildStates(states, projects)

    @simple_layout('layouts/repo-state-minimal.yaml')
    def test_repo_state_minimal(self):
        # Test that we don't restore unecessary repos
        self.executor_server.hold_jobs_in_build = True

        files = {'test.yaml': 'foo: bar'}
        self.addCommitToRepo('org/project3', 'add vars', files)

        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        build = self.getBuildByName('testjob1')
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project1')
        self.assertTrue(os.path.exists(path))
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project2')
        self.assertTrue(os.path.exists(path))
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project3')
        self.assertTrue(os.path.exists(path))

        build = self.getBuildByName('testjob2')
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project1')
        self.assertTrue(os.path.exists(path))
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project2')
        self.assertFalse(os.path.exists(path))
        path = os.path.join(build.jobdir.src_root,
                            'review.example.com', 'org', 'project3')
        self.assertFalse(os.path.exists(path))

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_repo_state_file(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        build = self.getBuildByName('project-merge')
        path = os.path.join(build.jobdir.log_root, 'workspace-repos.json')
        with open(path, 'r') as f:
            data = json.load(f)
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

        # Don't check the timestamp, but verify it exists
        self.assertIsNotNone(data['merge_ops'][2].pop('timestamp', None))
        self.assertEqual(
            data['merge_ops'],
            [{'cmd': ['git', 'checkout', 'master'],
              'path': 'review.example.com/org/project1'},
             {'cmd': ['git', 'fetch', 'origin', 'refs/changes/01/1/1'],
              'path': 'review.example.com/org/project1'},
             {'cmd': ['git',
                      'merge',
                      '-m',
                      "Merge 'refs/changes/01/1/1'",
                      '-s',
                      'resolve',
                      'FETCH_HEAD'],
              'path': 'review.example.com/org/project1'},
             {'cmd': ['git', 'checkout', 'master'],
              'path': 'review.example.com/org/project1'}])
        self.assertEqual(
            set(data['repo_state']['review.example.com/org/project1'].keys()),
            {'refs/heads/master', 'refs/remotes/origin/master',
             'refs/tags/init'})
        self.assertEqual(
            "zuul@example.com",
            data['merge_email'])
        self.assertEqual(
            "zuul",
            data['merge_name'])


class TestExecutorRepoRoles(ZuulTestCase, ExecutorReposMixin):
    tenant_config_file = 'config/executor-repos/main.yaml'

    def test_repo_checkout_roles_branch(self):
        self.executor_server.hold_jobs_in_build = True
        p0 = "review.example.com/org/project"
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p0, p1, p2]

        self.create_branch('org/project1', 'test-branch')
        self.create_branch('org/project2', 'test-branch')

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files={'branch': 'run branch'})
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        states = [
            {
                p0: dict(present=[A], branch='master'),
                p1: dict(absent=[A], branch='test-branch'),
                p2: dict(absent=[A], branch='test-branch'),
            },
        ]

        self.assertBuildStates(states, projects)

    def test_repo_checkout_roles_tag(self):
        self.executor_server.hold_jobs_in_build = True
        p0 = "review.example.com/org/project"
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p0, p1, p2]
        upstream = self.getUpstreamRepos(projects)

        self.create_branch('org/project1', 'unused-branch')
        files = {'README': 'tagged readme'}
        self.addCommitToRepo('org/project1', 'tagged commit',
                             files, branch='unused-branch', tag='test-tag')
        self.create_branch('org/project2', 'unused-branch')
        files = {'README': 'tagged readme'}
        self.addCommitToRepo('org/project2', 'tagged commit',
                             files, branch='unused-branch', tag='test-tag')

        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files={'tag': 'run tag'})
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        states = [
            {
                p0: dict(present=[A], branch='master'),
                p1: dict(commit=str(upstream[p1].commit('test-tag')),
                         absent=[A]),
                p2: dict(commit=str(upstream[p2].commit('test-tag')),
                         absent=[A]),
            },
        ]

        self.assertBuildStates(states, projects)


class TestAnsibleJob(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def run_job(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        return list(self.executor_server.job_workers.values())[0]

    def test_host_keys(self):
        self.fake_nodepool.host_keys = ['fake-host-key']
        job = self.run_job()
        keys = job.host_list[0]['host_keys']
        self.assertEqual(keys[0], '127.0.0.1 fake-host-key')

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_host_keys_connection_port(self):
        # Test with custom connection_port set
        self.fake_nodepool.host_keys = ['fake-host-key']
        self.fake_nodepool.connection_port = 22022
        job = self.run_job()
        keys = job.host_list[0]['host_keys']
        self.assertEqual(keys[0], '[127.0.0.1]:22022 fake-host-key')

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_no_host_keys_connection_port(self):
        # Test with no host keys
        self.fake_nodepool.host_keys = []
        self.fake_nodepool.connection_port = 22022
        job = self.run_job()
        keys = job.host_list[0]['host_keys']
        self.assertEqual(keys, [])
        self.assertEqual(
            job.host_list[0]['host_vars']['ansible_ssh_common_args'],
            '-o StrictHostKeyChecking=false')

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_no_shell_type(self):
        # Test without shell type set
        job = self.run_job()
        host = job.host_list[0]

        self.assertNotIn('ansible_shell_type', host['host_vars'])

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()

    def test_custom_shell_type(self):
        # Test with custom shell type set.
        self.fake_nodepool.shell_type = 'cmd'
        job = self.run_job()
        host = job.host_list[0]

        self.assertIn('ansible_shell_type', host['host_vars'])
        self.assertEqual(
            host['host_vars']['ansible_shell_type'],
            'cmd')

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()


class TestExecutorHostname(ZuulTestCase):
    config_file = 'zuul-executor-hostname.conf'
    tenant_config_file = 'config/single-tenant/main.yaml'

    def test_executor_hostname(self):
        self.assertEqual('test-executor-hostname.example.com',
                         self.executor_server.hostname)


class TestStartingBuildsSensor(ZuulTestCase):
    config_file = 'zuul.conf'
    tenant_config_file = 'config/governor/main.yaml'

    def test_default_case(self):
        # Given
        cores = multiprocessing.cpu_count()

        # When
        sensor = StartingBuildsSensor(None, '', None, cores * 2.5, None)

        # Then
        coefficient = 2 if multiprocessing.cpu_count() <= 4 else 1
        max_default = int(cores * 2.5 * coefficient)
        self.assertEqual(sensor.max_starting_builds, max_default)
        self.assertEqual(sensor.min_starting_builds, max(int(cores / 2), 1))

    def test_configuration_not_exists(self):
        # Given
        cores = multiprocessing.cpu_count()

        # When
        sensor = StartingBuildsSensor(None, '', None, cores * 2.5, self.config)

        # Then
        coefficient = 2 if multiprocessing.cpu_count() <= 4 else 1
        max_default = int(cores * 2.5 * coefficient)
        self.assertEqual(sensor.max_starting_builds, max_default)
        self.assertEqual(sensor.min_starting_builds, max(int(cores / 2), 1))

    def test_configuration_override(self):
        # Given
        cores = multiprocessing.cpu_count()
        self.config.set('executor', 'max_starting_builds', '5')

        # When
        sensor = StartingBuildsSensor(None, '', None, cores * 2.5, self.config)

        # Then
        self.assertEqual(sensor.max_starting_builds, 5)
        self.assertEqual(sensor.min_starting_builds, min(
            max(int(cores / 2), 1), sensor.max_starting_builds))

    def test_configuration_override_affecting_min(self):
        # Given
        cores = multiprocessing.cpu_count()
        self.config.set('executor', 'max_starting_builds', '1')

        # When
        sensor = StartingBuildsSensor(None, '', None, cores * 2.5, self.config)

        # Then
        self.assertEqual(sensor.max_starting_builds, 1)
        self.assertEqual(sensor.min_starting_builds, 1)


class TestGovernor(ZuulTestCase):
    config_file = 'zuul-executor-governor.conf'
    tenant_config_file = 'config/governor/main.yaml'

    @mock.patch('os.getloadavg')
    @mock.patch('psutil.virtual_memory')
    def test_load_governor(self, vm_mock, loadavg_mock):
        class Dummy(object):
            pass
        ram = Dummy()
        ram.percent = 20.0  # 20% used
        ram.total = 8 * 1024 * 1024 * 1024  # 8GiB
        vm_mock.return_value = ram
        loadavg_mock.return_value = (0.0, 0.0, 0.0)
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # fake the load to be higher than permitted
        fake_load = multiprocessing.cpu_count() * 2.6
        loadavg_mock.return_value = (fake_load, fake_load, fake_load)
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

    @mock.patch('os.getloadavg')
    @mock.patch('psutil.virtual_memory')
    def test_ram_governor(self, vm_mock, loadavg_mock):
        class Dummy(object):
            pass
        ram = Dummy()
        ram.percent = 20.0  # 20% used
        ram.total = 8 * 1024 * 1024 * 1024  # 8GiB
        vm_mock.return_value = ram
        loadavg_mock.return_value = (0.0, 0.0, 0.0)
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)
        ram.percent = 99.0  # 99% used
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

    @mock.patch('os.getloadavg')
    @mock.patch('psutil.virtual_memory')
    def test_ram_cgroup_governor(self, vm_mock, loadavg_mock):
        class Dummy(object):
            pass
        ram = Dummy()
        ram.percent = 20.0  # 20% used
        ram.total = 8 * 1024 * 1024 * 1024  # 8GiB
        vm_mock.return_value = ram
        loadavg_mock.return_value = (0.0, 0.0, 0.0)

        # Set no cgroup limit
        ram_sensor = [x for x in self.executor_server.sensors
                      if isinstance(x, RAMSensor)][0]
        ram_sensor.cgroup_stats_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'memory.stat.nolimit')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Set cgroup limit 5GiB and ram usage 20%
        ram_sensor.cgroup_stats_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'memory.stat.ok')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Set cgroup limit 5GiB and ram usage 96%
        ram_sensor.cgroup_stats_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'memory.stat.bad')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

    @mock.patch('os.getloadavg')
    @mock.patch('psutil.virtual_memory')
    @okay_tracebacks("invalid literal for int() with base 10: 'foo'")
    def test_process_governor(self, vm_mock, loadavg_mock):
        # Set up load average and memory sensors to accept work.
        class Dummy(object):
            pass
        ram = Dummy()
        ram.percent = 20.0  # 20% used
        ram.total = 8 * 1024 * 1024 * 1024  # 8GiB
        vm_mock.return_value = ram
        loadavg_mock.return_value = (0.0, 0.0, 0.0)

        # Test no limit
        process_sensor = [x for x in self.executor_server.sensors
                          if isinstance(x, ProcessSensor)][0]
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.max')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.max')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test typical runtime values
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.8k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.8k')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test within 5%
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.15600')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.15600')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        # Test within 10 keep running
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.150')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.100')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.150')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.100')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test within 10 stop
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.150')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.141')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.150')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.141')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        # Test at limit
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        # Test no limit only root cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.max')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test typical runtime values only root cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.8k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test at limit only root cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        # Test no limit only user cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.max')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test typical runtime values only user cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.8k')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test at limit only user cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.16k')
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        # Test no cgroup files found
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'doesnotexist')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        # Test inability to parse cgroup files
        process_sensor._root_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.foo')
        process_sensor._root_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.foo')
        process_sensor._user_cgroup_max_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.foo')
        process_sensor._user_cgroup_cur_file = os.path.join(
            FIXTURE_DIR, 'cgroup', 'pids.foo')
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

    @mock.patch('os.getloadavg')
    @mock.patch('os.statvfs')
    def test_hdd_governor(self, statvfs_mock, loadavg_mock):
        class Dummy(object):
            pass
        hdd = Dummy()
        hdd.f_frsize = 4096
        hdd.f_blocks = 120920708
        hdd.f_bfree = 95716701  # 20.84% used
        hdd.f_files = 61022208
        hdd.f_ffree = 32147841  # 47.31% used
        statvfs_mock.return_value = hdd
        loadavg_mock.return_value = (0.0, 0.0, 0.0)

        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_hdd',
            value='2084', kind='g')
        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_inodes',
            value='4731', kind='g')

        hdd.f_bfree = 5716701  # 95.27% used
        statvfs_mock.return_value = hdd

        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_hdd',
            value='9527', kind='g')
        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_inodes',
            value='4731', kind='g')

        hdd.f_bfree = 95716701  # 20.84% used
        hdd.f_ffree = 1336387  # 97.80% used
        statvfs_mock.return_value = hdd

        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_hdd',
            value='2084', kind='g')
        self.assertReportedStat(
            'zuul.executor.test-executor-hostname_example_com.pct_used_inodes',
            value='9780', kind='g')

    @mock.patch('os.getloadavg')
    def test_pause_governor(self, loadavg_mock):
        loadavg_mock.return_value = (0.0, 0.0, 0.0)

        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)

        self.executor_server.pause_sensor.pause = True
        self.executor_server.manageLoad()
        self.assertFalse(self.executor_server.accepting_work)

    def waitForExecutorBuild(self, jobname):
        self.log.debug("Waiting for %s to start", jobname)
        timeout = time.time() + 30
        build = None
        while time.time() < timeout and not build:
            for b in self.builds:
                if b.name == jobname:
                    build = b
                    break
            time.sleep(0.1)
        self.log.debug("Found build %s", jobname)
        build_id = build.uuid
        while (time.time() < timeout and
               build_id not in self.executor_server.job_workers):
            time.sleep(0.1)
        worker = self.executor_server.job_workers[build_id]
        self.log.debug("Found worker %s", jobname)
        while (time.time() < timeout and
               not worker.started):
            time.sleep(0.1)
        self.log.debug("Worker for %s started: %s", jobname, worker.started)
        return build

    @mock.patch('os.getloadavg')
    def test_slow_start(self, loadavg_mock):
        loadavg_mock.return_value = (0.0, 0.0, 0.0)

        def _set_starting_builds(min, max):
            for sensor in self.executor_server.sensors:
                if isinstance(sensor, StartingBuildsSensor):
                    sensor.min_starting_builds = min
                    sensor.max_starting_builds = max

        # Note: This test relies on the fact that manageLoad is only
        # run at specific points.  Several times in the test we check
        # that manageLoad has disabled or enabled job acceptance based
        # on the number of "starting" jobs.  Some of those jobs may
        # have actually moved past the "starting" phase and are
        # actually "running".  But because manageLoad hasn't run
        # again, it still uses the old values.  Keep this in mind when
        # double checking its calculations.
        #
        # Disable the periodic governor runs to make sure they don't
        # interefere (only possible if the test runs longer than 10
        # seconds).
        self.executor_server.governor_stop_event.set()
        self.executor_server.hold_jobs_in_build = True
        _set_starting_builds(1, 1)
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)
        A = self.fake_gerrit.addFakeChange('common-config', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))

        build1 = self.waitForExecutorBuild('test1')
        # With one job (test1) being started, we should no longer
        # be accepting new work
        self.assertFalse(self.executor_server.accepting_work)
        self.assertEqual(len(self.executor_server.job_workers), 1)
        # Allow enough starting builds for the test to complete.
        _set_starting_builds(1, 3)
        # We must wait for build1 to enter a waiting state otherwise
        # the subsequent release() is a noop and the build is never
        # released.  We don't use waitUntilSettled as that requires
        # the other two builds to start which can't happen while we
        # don't accept jobs.
        for x in iterate_timeout(30, "build1 is waiting"):
            if build1.waiting:
                break
        build1.release()
        for x in iterate_timeout(30, "Wait for build1 to complete"):
            if build1.uuid not in self.executor_server.job_workers:
                break
        self.executor_server.manageLoad()
        # This manageLoad call has determined that there are 0 workers
        # running, so our full complement of 3 starting builds is
        # available.  It will re-register for work and pick up the
        # next two jobs.

        self.waitForExecutorBuild('test2')
        self.waitForExecutorBuild('test3')
        # When each of these jobs started, they caused manageLoad to
        # be called, the second invocation calculated that there were
        # 2 workers running, so our starting build limit was reduced
        # to 1.  Usually it will calculate that there are 2 starting
        # builds, but theoretically it could count only 1 if the first
        # build manages to leave the starting phase before the second
        # build starts.  It should always count the second build as
        # starting.  As long as at least one build is still in the
        # starting phase, this will exceed the limit and unregister.
        self.assertFalse(self.executor_server.accepting_work)

        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.executor_server.manageLoad()
        self.assertTrue(self.executor_server.accepting_work)


class TestLineMapping(AnsibleZuulTestCase):
    config_file = 'zuul-gerrit-web.conf'
    tenant_config_file = 'config/line-mapping/main.yaml'

    def _test_line_mapping(self, method):
        with open(os.path.join(FIXTURE_DIR,
                               'config/line-mapping/git/',
                               'org_project/README')) as f:
            content = f.read().split('\n')

        # The change under test adds a line to the end.
        change_content = content + ['end line']
        file_dict = {'README': '\n'.join(change_content)}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        if method == 'merge':
            pass
        elif method == 'cherry_pick':
            A.cherry_pick = True
        else:
            raise Exception("unsupported method")

        # An intervening change adds a line near the top.
        # And changes a line in the middle (no net offset from that),
        # just to make sure we have more than one diff hunk.
        head_content = (
            content[:4] +
            ['new line 5'] +
            content[4:50] +
            ['changed line 51'] +
            content[51:]
        )
        file_dict = {'README': '\n'.join(head_content)}
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B',
                                           files=file_dict)
        B.setMerged()

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('file-comments').result,
                         'SUCCESS')
        self.assertEqual(len(A.comments), 5)
        comments = sorted(A.comments, key=lambda x: x['line'])
        reviewer = {'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'}
        self.assertEqual(comments[0], {
            'file': 'README',
            'line': 3,
            'message': 'line 3 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[1], {
            'file': 'README',
            'line': 7,
            'message': 'line 7 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[2], {
            "file": "README",
            "line": 14,
            "message": "range comment for line 14",
            "range": {
                "end_character": 26,
                "end_line": 14,
                "start_character": 0,
                "start_line": 11,
            },
            "reviewer": {
                "email": "zuul@example.com",
                "name": "Zuul",
                "username": "jenkins",
            }
        })
        self.assertEqual(comments[3], {
            'file': 'README',
            'line': 100,
            'message': 'line 100 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[4], {
            'file': 'README',
            'line': 101,
            'message': 'line 101 comment',
            'reviewer': reviewer,
        })

    def test_line_mapping_merge(self):
        self._test_line_mapping(method='merge')

    def test_line_mapping_cherry_pick(self):
        self._test_line_mapping(method='cherry_pick')

    def test_line_mapping_fast_forward(self):
        with open(os.path.join(FIXTURE_DIR,
                               'config/line-mapping/git/',
                               'org_project/README')) as f:
            content = f.read().split('\n')

        # The change under test adds a line to the end and also the middle.
        change_content = (content[:4] + ['new line 5'] +
                          content[4:] + ['end line'])
        file_dict = {'README': '\n'.join(change_content)}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)

        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(self.getJobFromHistory('file-comments').result,
                         'SUCCESS')
        self.assertEqual(len(A.comments), 5)
        comments = sorted(A.comments, key=lambda x: x['line'])
        reviewer = {'email': 'zuul@example.com',
                    'name': 'Zuul',
                    'username': 'jenkins'}
        self.assertEqual(comments[0], {
            'file': 'README',
            'line': 3,
            'message': 'line 3 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[1], {
            'file': 'README',
            'line': 8,
            'message': 'line 7 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[2], {
            "file": "README",
            "line": 15,
            "message": "range comment for line 14",
            "range": {
                "end_character": 26,
                "end_line": 15,
                "start_character": 0,
                "start_line": 12,
            },
            "reviewer": {
                "email": "zuul@example.com",
                "name": "Zuul",
                "username": "jenkins",
            }
        })
        self.assertEqual(comments[3], {
            'file': 'README',
            'line': 101,
            'message': 'line 100 comment',
            'reviewer': reviewer,
        })
        self.assertEqual(comments[4], {
            'file': 'README',
            'line': 102,
            'message': 'line 101 comment',
            'reviewer': reviewer,
        })


class ExecutorFactsMixin:
    # These should be overridden in child classes.
    tenant_config_file = 'config/executor-facts/main.yaml'
    ansible_major_minor = 'X.Y'

    def _get_file(self, build, path):
        p = os.path.join(build.jobdir.root, path)
        with open(p) as f:
            return f.read()

    def test_datetime_fact(self):
        self.executor_server.keep_jobdir = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(self.getJobFromHistory('datetime-fact').result,
                         'SUCCESS')

        j = json.loads(self._get_file(self.history[0],
                                      'work/logs/job-output.json'))

        date_time = \
            j[0]['plays'][0]['tasks'][0]['hosts']['localhost']['date_time']
        self.assertEqual(18, len(date_time))
        build = self.getJobFromHistory('datetime-fact', result='SUCCESS')
        with open(build.jobdir.job_output_file) as f:
            output = f.read()
            self.assertIn(f'Ansible version={self.ansible_major_minor}',
                          output)


class TestExecutorFacts9(AnsibleZuulTestCase, ExecutorFactsMixin):
    tenant_config_file = 'config/executor-facts/main9.yaml'
    ansible_major_minor = '2.16'


class TestExecutorFacts11(AnsibleZuulTestCase, ExecutorFactsMixin):
    tenant_config_file = 'config/executor-facts/main11.yaml'
    ansible_major_minor = '2.18'


class AnsibleCallbackConfigsMixin:
    config_file = 'zuul-executor-ansible-callback.conf'

    # These should be overridden in child classes.
    tenant_config_file = 'config/ansible-callbacks/main.yaml'
    ansible_major_minor = 'X.Y'

    def test_ansible_callback_config(self):
        self.executor_server.keep_jobdir = True
        A = self.fake_gerrit.addFakeChange('common-config', 'master', 'A')
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        callbacks = [
            'callback_test_callback',
            'callback_nocows = True',
            'callback_nocows = False',
            'callback_\\nnocows = True',
            'callback_\\nnocows = False',
            'callback_ansible_interpolation'
        ]

        p = os.path.join(self.getJobFromHistory('callback-test').jobdir.root,
                         'ansible/playbook_0/ansible.cfg')
        self.assertEqual(self.getJobFromHistory('callback-test').result,
                         'SUCCESS')

        c = configparser.ConfigParser(interpolation=None)
        c.read(p)
        for callback in callbacks:
            self.assertIn(callback, c.sections())
        self.assertIn('test_field', c['callback_ansible_interpolation'])
        self.assertIn('test-%-value',
                      c['callback_ansible_interpolation']['test_field'])

        self.assertIn('file_name', c['callback_test_callback'])
        self.assertEqual('callback-success',
                         c['callback_test_callback']['file_name'])
        callback_result_file = os.path.join(
            self.getJobFromHistory('callback-test').jobdir.root,
            'trusted/project_0/review.example.com/',
            'common-config/playbooks/callback_plugins/',
            c['callback_test_callback']['file_name'])
        self.assertTrue(os.path.isfile(callback_result_file))
        build = self.getJobFromHistory('callback-test', result='SUCCESS')
        with open(build.jobdir.job_output_file) as f:
            output = f.read()
            self.assertIn(f'Ansible version={self.ansible_major_minor}',
                          output)


class TestAnsibleCallbackConfigs9(AnsibleZuulTestCase,
                                  AnsibleCallbackConfigsMixin):
    config_file = 'zuul-executor-ansible-callback.conf'
    tenant_config_file = 'config/ansible-callbacks/main9.yaml'
    ansible_major_minor = '2.16'


class TestAnsibleCallbackConfigs11(AnsibleZuulTestCase,
                                   AnsibleCallbackConfigsMixin):
    config_file = 'zuul-executor-ansible-callback.conf'
    tenant_config_file = 'config/ansible-callbacks/main11.yaml'
    ansible_major_minor = '2.18'


class TestExecutorEnvironment(AnsibleZuulTestCase):
    tenant_config_file = 'config/zuul-environment-filter/main.yaml'

    @mock.patch.dict('os.environ', {'ZUUL_TEST_VAR': 'some-value',
                                    'TEST_VAR': 'not-empty'})
    def test_zuul_environment_filter(self):
        A = self.fake_gerrit.addFakeChange('common-config', 'master', 'A')
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()

        self.assertEqual(
            self.getJobFromHistory('zuul-environment-filter').result,
            'SUCCESS')


class TestExecutorStart(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    def setup_config(self, config_file: str):
        config = super(TestExecutorStart, self).setup_config(config_file)
        self.junk_dir = os.path.join(self.jobdir_root, 'junk')
        os.makedirs(self.junk_dir)
        return config

    def test_executor_start(self):
        self.assertFalse(os.path.exists(self.junk_dir))


class TestExecutorExtraPackages(AnsibleZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    test_package = 'pywinrm'

    def setUp(self):
        super(TestExecutorExtraPackages, self).setUp()
        import subprocess
        ansible_manager = self.executor_server.ansible_manager
        for version in ansible_manager._supported_versions:
            command = [ansible_manager.getAnsibleCommand(version, 'pip'),
                       'uninstall', '-y', self.test_package]
            # We want to error if the uninstall fails as the test below
            # relies on the package not being installed to be properly
            # exercised.
            s = subprocess.run(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
            self.log.info(s.stdout)
            self.assertEqual(s.returncode, 0)

    @mock.patch('zuul.lib.ansible.ManagedAnsible.extra_packages',
                new_callable=mock.PropertyMock)
    def test_extra_packages(self, mock_extra_packages):
        mock_extra_packages.return_value = [self.test_package]
        ansible_manager = self.executor_server.ansible_manager
        self.assertFalse(ansible_manager.validate())
        ansible_manager.install()
        self.assertTrue(ansible_manager.validate())


class TestVarSquash(BaseTestCase):
    def test_squash_variables(self):
        # Test that we correctly squash job variables
        nodeset = NodeSet()
        nodes = [
            {'name': 'node1', 'host_vars': {
                'host': 'node1_host',
                'extra': 'node1_extra',
            }},
            {'name': 'node2', 'host_vars': {
                'host': 'node2_host',
                'extra': 'node2_extra',
            }},
        ]
        nodeset.addGroup(Group('group1', ['node1']))
        nodeset.addGroup(Group('group2', ['node2']))
        groupvars = {
            'group1': {
                'host': 'group1_host',
                'group': 'group1_group',
                'extra': 'group1_extra',
            },
            'group2': {
                'host': 'group2_host',
                'group': 'group2_group',
                'extra': 'group2_extra',
            },
            'all': {
                'all2': 'groupvar_all2',
            }
        }
        jobvars = {
            'host': 'jobvar_host',
            'group': 'jobvar_group',
            'all': 'jobvar_all',
            'extra': 'jobvar_extra',
        }
        extravars = {
            'extra': 'extravar_extra',
        }
        out = squash_variables(
            nodes, nodeset, jobvars, groupvars, extravars)

        expected = {
            'node1': {
                'all': 'jobvar_all',
                'all2': 'groupvar_all2',
                'group': 'group1_group',
                'host': 'node1_host',
                'extra': 'extravar_extra'},
            'node2': {
                'all': 'jobvar_all',
                'all2': 'groupvar_all2',
                'group': 'group2_group',
                'host': 'node2_host',
                'extra': 'extravar_extra'},
        }
        self.assertEqual(out, expected)


class TestExecutorFailure(ZuulTestCase):
    tenant_config_file = 'config/single-tenant/main.yaml'

    @mock.patch('zuul.executor.server.ExecutorServer.executeJob')
    @okay_tracebacks('Failed to start')
    def test_executor_job_start_failure(self, execute_job_mock):
        execute_job_mock.side_effect = Exception('Failed to start')
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertTrue(re.search(
            '- project-merge .* ERROR',
            A.messages[-1]))

    @okay_tracebacks("Transient error")
    def test_executor_transient_error(self):
        self.hold_jobs_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A')
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()

        with mock.patch('zuul.merger.merger.Merger.updateRepo') as update_mock:
            update_mock.side_effect = IOError("Transient error")
            self.executor_api.release()
            self.waitUntilSettled()

        items = self.getAllItems('tenant-one', 'gate')
        self.assertEqual(len(items), 1)

        self.hold_jobs_in_queue = False
        self.executor_api.release()
        self.waitUntilSettled()

        job = items[0].current_build_set.job_graph.getJob(
            'project-merge', items[0].changes[0].cache_key)
        build_retries = items[0].current_build_set.getRetryBuildsForJob(job)
        self.assertEqual(len(build_retries), 1)
        self.assertIsNotNone(build_retries[0].error_detail)
        self.assertTrue(
            build_retries[0].error_detail.startswith(
                "Failed to update project"))


class TestExecutorWorkspaceCheckout(ZuulTestCase, ExecutorReposMixin):
    tenant_config_file = 'config/workspace-checkout/main.yaml'

    def _test_workspace_checkout(self, jobname):
        self.executor_server.keep_jobdir = True
        self.executor_server.hold_jobs_in_build = True
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p1, p2]
        upstream = self.getUpstreamRepos(projects)

        files = {jobname: 'run'}
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A',
                                           files=files)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        states = [
            {p1: dict(present=[A]),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      absent=[A]),
             },
        ]

        self.assertBuildStates(states, projects)

    def _assert_not_sparse(self):
        cc = "review.example.com/common-config"
        rt = "review.example.com/org/roletest"
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        path = os.path.join(self.history[0].jobdir.root)
        file_checks = {
            f'trusted/project_0/{cc}/otherdirectory/something.txt': True,
            f'trusted/project_0/{cc}/playbooks/run.yaml': True,
            f'trusted/project_0/{cc}/includevars/vars.yaml': True,
            f'trusted/project_1/{rt}/otherdirectory/something.txt': True,
            f'trusted/project_1/{rt}/tasks/main.yaml': True,
            f'work/src/{p1}/otherdirectory/something.txt': True,
            f'work/src/{p2}/otherdirectory/something.txt': True,
        }
        for f, expected in file_checks.items():
            exists = os.path.exists(os.path.join(path, f))
            self.assertEqual(expected, exists, f)

    def _assert_sparse(self):
        cc = "review.example.com/common-config"
        rt = "review.example.com/org/roletest"
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        path = os.path.join(self.history[0].jobdir.root)
        file_checks = {
            f'trusted/project_0/{cc}/otherdirectory/something.txt': False,
            f'trusted/project_0/{cc}/playbooks/run.yaml': True,
            f'trusted/project_0/{cc}/includevars/vars.yaml': True,
            f'trusted/project_1/{rt}/otherdirectory/something.txt': False,
            f'trusted/project_1/{rt}/tasks/main.yaml': True,
            f'work/src/{p1}/otherdirectory/something.txt': False,
            f'work/src/{p2}/otherdirectory/something.txt': False,
        }
        for f, expected in file_checks.items():
            exists = os.path.exists(os.path.join(path, f))
            self.assertEqual(expected, exists, f)

    def test_workspace_checkout_true(self):
        self._test_workspace_checkout('test-workspace-checkout-true')
        self._assert_not_sparse()

    def test_workspace_checkout_false(self):
        self._test_workspace_checkout('test-workspace-checkout-false')
        self._assert_sparse()

    def test_workspace_checkout_auto_nodeset(self):
        self._test_workspace_checkout(
            'test-workspace-checkout-auto-nodeset')
        self._assert_sparse()

    def test_workspace_checkout_auto_no_nodeset(self):
        self._test_workspace_checkout(
            'test-workspace-checkout-auto-no-nodeset')
        self._assert_not_sparse()

    def test_workspace_checkout_line_mapping(self):
        self.executor_server.keep_jobdir = True
        self.executor_server.hold_jobs_in_build = True
        cc = "review.example.com/common-config"
        rt = "review.example.com/org/roletest"
        p1 = "review.example.com/org/project1"
        p2 = "review.example.com/org/project2"
        projects = [p1, p2]
        upstream = self.getUpstreamRepos(projects)

        jobname = 'test-workspace-checkout-false'
        files = {jobname: 'run'}
        A = self.fake_gerrit.addFakeChange('org/project1', 'master', 'A',
                                           files=files)
        self.executor_server.returnData(
            jobname, A,
            {'zuul':
             {'file_comments':
              {'otherdirectory/something.txt': [{
                  'line': 1,
                  'message': 'test message'}]
               }
              }
             }
        )
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        states = [
            {p1: dict(present=[A]),
             p2: dict(commit=str(upstream[p2].commit('master')),
                      absent=[A]),
             },
        ]

        self.assertBuildStates(states, projects)
        path = os.path.join(self.history[0].jobdir.root)
        file_checks = {
            f'trusted/project_0/{cc}/otherdirectory/something.txt': False,
            f'trusted/project_0/{cc}/playbooks/run.yaml': True,
            f'trusted/project_0/{cc}/includevars/vars.yaml': True,
            f'trusted/project_1/{rt}/otherdirectory/something.txt': False,
            f'trusted/project_1/{rt}/tasks/main.yaml': True,
            f'work/src/{p1}/otherdirectory/something.txt': False,
            f'work/src/{p2}/otherdirectory/something.txt': False,
        }
        for f, expected in file_checks.items():
            exists = os.path.exists(os.path.join(path, f))
            self.assertEqual(expected, exists, f)

    def test_workspace_checkout_multi_include_vars(self):
        A = self.fake_gerrit.addFakeChange('org/project2', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        self.assertHistory([
            dict(
                name='test-workspace-checkout-multi-include-vars',
                result='SUCCESS'
            ),
        ], ordered=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add build final column

Revision ID: 269691d2220e
Revises: e0eda5d09eae
Create Date: 2020-01-03 07:53:15.962739

"""

# revision identifiers, used by Alembic.
revision = '269691d2220e'
down_revision = '16c1dc9054d0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILD_TABLE = 'zuul_build'


def upgrade(table_prefix=''):
    op.add_column(table_prefix + BUILD_TABLE, sa.Column('final', sa.Boolean))

    # Set all existing build entries to final (otherwise they will vanish from
    # the UI)
    new_column = sa.table(table_prefix + BUILD_TABLE,
                          sa.Column('final', sa.Boolean))
    op.execute(new_column.update().values(**{'final': True}))


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import time

import fixtures
import testtools
from kazoo.exceptions import NoNodeError

from zuul import model
from zuul.launcher.client import LauncherClient

from tests.base import (
    ZuulTestCase,
    iterate_timeout,
)


class BaseCloudDriverTest(ZuulTestCase):
    cloud_test_connection_type = 'ssh'
    cloud_test_image_format = ''
    cloud_test_provider_name = ''
    # The number of instances to expect when nothing is running
    cloud_test_min_instances = 0

    def setUp(self):
        self.useFixture(fixtures.MonkeyPatch(
            'zuul.launcher.server.NodescanRequest.FAKE', True))
        super().setUp()

    def _getProvider(self):
        # Use the launcher provider so that we're using the same ttl
        # method caches.
        for provider in self.launcher.tenant_providers['tenant-one']:
            if provider.name == self.cloud_test_provider_name:
                return provider

    def _assertProviderNodeAttributes(self, pnode):
        self.assertEqual(pnode.connection_type,
                         self.cloud_test_connection_type)
        self.assertIsNotNone(pnode.interface_ip)

    def _test_node_lifecycle(self, label):
        # Call this in a test to run a node lifecycle
        for _ in iterate_timeout(
                30, "scheduler and launcher to have the same layout"):
            if (self.scheds.first.sched.local_layout_state.get("tenant-one") ==
                self.launcher.local_layout_state.get("tenant-one")):
                break
        provider = self._getProvider()
        nodeset = model.NodeSet()
        nodeset.addNode(model.Node("node", label))

        ctx = self.createZKContext(None)
        request = self.requestNodes([n.label for n in nodeset.getNodes()])

        client = LauncherClient(self.zk_client, None)
        request = client.getRequest(request.uuid)

        self.assertEqual(request.state, model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request.nodes), 1)

        client.acceptNodeset(request, nodeset)
        self.waitUntilSettled()

        with testtools.ExpectedException(NoNodeError):
            # Request should be gone
            request.refresh(ctx)

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertIsNotNone(pnode)
            self.assertTrue(pnode.hasLock())
            self._assertProviderNodeAttributes(pnode)

        for _ in iterate_timeout(10, "instances to appear"):
            if (len(list(provider.endpoint.listInstances())) >
                self.cloud_test_min_instances):
                break
        client.useNodeset(nodeset)
        self.waitUntilSettled()

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertTrue(pnode.hasLock())
            self.assertEqual(pnode.state, pnode.State.IN_USE)

        client.returnNodeset(nodeset)
        self.waitUntilSettled()

        for node in nodeset.getNodes():
            pnode = node._provider_node
            self.assertFalse(pnode.hasLock())
            self.assertEqual(pnode.state, pnode.State.USED)

            for _ in iterate_timeout(60, "node to be deleted"):
                try:
                    pnode.refresh(ctx)
                except NoNodeError:
                    break

        # Iterate here because the aws driver (at least) performs
        # delayed async deletes.
        for _ in iterate_timeout(60, "instances to be deleted"):
            if (len(list(provider.endpoint.listInstances())) ==
                self.cloud_test_min_instances):
                break

    def _test_quota(self, label):
        for _ in iterate_timeout(
                30, "scheduler and launcher to have the same layout"):
            if (self.scheds.first.sched.local_layout_state.get("tenant-one") ==
                self.launcher.local_layout_state.get("tenant-one")):
                break
        provider = self._getProvider()
        nodeset1 = model.NodeSet()
        nodeset1.addNode(model.Node("node", label))

        nodeset2 = model.NodeSet()
        nodeset2.addNode(model.Node("node", label))

        ctx = self.createZKContext(None)
        request1 = self.requestNodes([n.label for n in nodeset1.getNodes()])
        request2 = self.requestNodes(
            [n.label for n in nodeset2.getNodes()],
            timeout=None)

        client = LauncherClient(self.zk_client, None)
        request1 = client.getRequest(request1.uuid)

        self.assertEqual(request1.state, model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request1.nodes), 1)

        client.acceptNodeset(request1, nodeset1)
        client.useNodeset(nodeset1)

        # We should still be waiting on request2.

        # TODO: This is potentially racy (but only producing
        # false-negatives) and also slow. We should find a way to
        # determine if the launcher had paused a provider.
        with testtools.ExpectedException(Exception):
            self.waitForNodeRequest(request2, 10)
        request2 = client.getRequest(request2.uuid)
        self.assertEqual(request2.state, model.NodesetRequest.State.REQUESTED)

        client.returnNodeset(nodeset1)
        self.waitUntilSettled()

        for node in nodeset1.getNodes():
            pnode = node._provider_node
            for _ in iterate_timeout(60, "node to be deleted"):
                try:
                    pnode.refresh(ctx)
                except NoNodeError:
                    break

        self.waitForNodeRequest(request2, 10)
        request2 = client.getRequest(request2.uuid)
        self.assertEqual(request2.state, model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request2.nodes), 1)

        client.acceptNodeset(request2, nodeset2)
        client.useNodeset(nodeset2)
        client.returnNodeset(nodeset2)
        self.waitUntilSettled()

        # Iterate here because the aws driver (at least) performs
        # delayed async deletes.
        for _ in iterate_timeout(60, "instances to be deleted"):
            if (len(list(provider.endpoint.listInstances())) ==
                self.cloud_test_min_instances):
                break

    def _test_diskimage(self, expected_uploads=1):
        self.addImageBuildEvent(
            'tenant-one',
            'review.example.com/org/common-config',
            'master',
            ['debian-local'],
        )
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='build-debian-local-image', result='SUCCESS'),
        ], ordered=False)

        name = 'review.example.com%2Forg%2Fcommon-config/debian-local'
        artifacts = self.launcher.image_build_registry.\
            getArtifactsForImage(name)
        self.assertEqual(1, len(artifacts))
        self.assertEqual(self.cloud_test_image_format, artifacts[0].format)
        self.assertTrue(artifacts[0].validated)
        uploads = self.launcher.image_upload_registry.getUploadsForImage(
            name)
        self.assertEqual(expected_uploads, len(uploads))
        for x in range(expected_uploads):
            self.assertEqual(artifacts[0].uuid, uploads[x].artifact_uuid)
            self.assertIsNotNone(uploads[x].external_id)
            self.assertTrue(uploads[x].validated)

    def _drive_state_machine(self, ctx, node, future_names, provider, create):
        executed_future_names = set()
        next_future_names = set()
        for _ in iterate_timeout(60, "create state machine to complete"):
            with node.activeContext(ctx):
                # Re-create the SM from the state in ZK
                if create:
                    sm = provider.getCreateStateMachine(
                        node, None, self.log)
                    node.create_state_machine = sm
                else:
                    sm = provider.getDeleteStateMachine(
                        node, self.log)
                    node.delete_state_machine = sm

                with self._block_futures():
                    sm.advance()
                # If there are pending futures we will try to re-create
                # the SM once from the state and then advance it once
                # more so the futures can complete.
                pending_futures = {
                    name: future
                    for name in future_names
                    if (future := getattr(sm, name, None))
                }
                this_futures = {
                    name: future for (name, future)
                    in pending_futures.items()
                    if name in next_future_names
                }
                if this_futures:
                    concurrent.futures.wait(this_futures.values())
                    executed_future_names.update(this_futures.keys())
                    sm.advance()
                next_future_names = (set(pending_futures.keys()) -
                                     executed_future_names)
            if sm.complete:
                break
            # Avoid busy-looping as we have to wait for the TTL
            # cache to expire.
            time.sleep(0.5)

    def _test_state_machines(self, label_name, provider_name,
                             node_class, future_names):
        # Stop the launcher main loop, so we can drive the state machine
        # on our own.
        self.waitUntilSettled()
        self.launcher._running = False
        self.launcher.wake_event.set()
        self.launcher.launcher_thread.join()

        layout = self.scheds.first.sched.abide.tenants.get('tenant-one').layout
        provider = layout.providers[provider_name]
        # Start the endpoint since we're going to use the scheduler's endpoint.
        provider.getEndpoint().start()

        with self.createZKContext(None) as ctx:
            node = node_class.new(ctx, label=label_name,
                                  uuid='1234',
                                  tags={'zuul_node_uuid': '1234'})
            self._drive_state_machine(ctx, node, future_names, provider, True)
            self._drive_state_machine(ctx, node, future_names, provider, False)



# Copyright 2018 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import os

from zuul.lib import strings
from zuul.zk.components import COMPONENT_REGISTRY


def make_src_dir(canonical_hostname, name, scheme):
    return os.path.join('src',
                        strings.workspace_project_path(
                            canonical_hostname,
                            name,
                            scheme))


def make_change(item, change):
    ret = change.toDict()
    if sha := item.merged_change_shas.get(change.cache_key):
        ret['merge_commit_id'] = sha
    return ret


def construct_build_params(uuid, connections, job, item, pipeline,
                           dependent_changes=[], merger_items=[],
                           redact_secrets_and_keys=True):
    """Returns a list of all the parameters needed to build a job.

    These parameters may be passed to zuul-executors (via ZK) to perform
    the job itself.

    Alternatively they contain enough information to load into another build
    environment - for example, a local runner.
    """
    tenant = item.manager.tenant
    change = item.getChangeForJob(job)
    project = dict(
        name=change.project.name,
        short_name=change.project.name.split('/')[-1],
        canonical_hostname=change.project.canonical_hostname,
        canonical_name=change.project.canonical_name,
        src_dir=make_src_dir(change.project.canonical_hostname,
                             change.project.name,
                             job.workspace_scheme),
    )

    # We override some values like project, so set the change fields
    # first.
    zuul_params = change.toDict()
    zuul_params.update(dict(
        build=uuid,
        buildset=item.current_build_set.uuid,
        ref=change.ref,
        buildset_refs=[make_change(item, c) for c in item.changes],
        build_refs=[make_change(item, c) for c in item.changes
                    if c.cache_key in job.all_refs],
        pipeline=pipeline.name,
        post_review=pipeline.post_review,
        job=job.name,
        project=project,
        tenant=tenant.name,
        event_id=item.event.zuul_event_id if item.event else None,
        jobtags=sorted(job.tags),
        include_vars=job.include_vars,
    ))
    if hasattr(change, 'message'):
        zuul_params['message'] = strings.b64encode(change.message)
    zuul_params['projects'] = {}  # Set below

    # Fixup the src_dir for the items based on this job
    dependent_changes = copy.deepcopy(dependent_changes)
    for dep_change in dependent_changes:
        dep_change['project']['src_dir'] = make_src_dir(
            dep_change['project']['canonical_hostname'],
            dep_change['project']['name'],
            job.workspace_scheme)
    # Fixup the src_dir for the refs based on this job
    for r in zuul_params['buildset_refs']:
        r['project']['src_dir'] = make_src_dir(
            r['project']['canonical_hostname'],
            r['project']['name'],
            job.workspace_scheme)
        # TODO: backwards compat, remove after zuul 14
        r['src_dir'] = make_src_dir(
            r['project']['canonical_hostname'],
            r['project']['name'],
            job.workspace_scheme)
    for r in zuul_params['build_refs']:
        r['project']['src_dir'] = make_src_dir(
            r['project']['canonical_hostname'],
            r['project']['name'],
            job.workspace_scheme)
        # TODO: backwards compat, remove after zuul 14
        r['src_dir'] = make_src_dir(
            r['project']['canonical_hostname'],
            r['project']['name'],
            job.workspace_scheme)

    zuul_params['items'] = dependent_changes
    zuul_params['child_jobs'] = [
        x.name for x in item.current_build_set.job_graph.
        getDirectDependentJobs(job)
    ]

    params = dict()
    (
        params["parent_data"],
        params["secret_parent_data"],
        artifact_data
    ) = item.getJobParentData(job)
    if artifact_data:
        zuul_params['artifacts'] = artifact_data

    params['job_ref'] = job.getPath()
    params['items'] = merger_items
    params['projects'] = []
    if hasattr(change, 'branch'):
        params['branch'] = change.branch
    else:
        params['branch'] = None
    params['repo_state_keys'] = item.current_build_set.repo_state_keys

    params['ssh_keys'] = []
    if COMPONENT_REGISTRY.model_api >= 36:
        if redact_secrets_and_keys:
            params['ssh_keys'].append("REDACTED")
        else:
            params['ssh_keys'].append(
                dict(tenant_name=tenant.name)
            )
    if pipeline.post_review:
        if redact_secrets_and_keys:
            params['ssh_keys'].append("REDACTED")
        else:
            params['ssh_keys'].append(dict(
                connection_name=change.project.connection_name,
                project_name=change.project.name))
    params['pipeline_builds_images'] = pipeline.builds_images
    params['zuul'] = zuul_params
    projects = set()
    required_projects = set()

    def make_project_dict(project, override_branch=None,
                          override_checkout=None):
        project_metadata = item.current_build_set.job_graph.\
            getProjectMetadata(project.canonical_name)
        if project_metadata:
            project_default_branch = project_metadata.default_branch
        else:
            project_default_branch = 'master'
        connection = project.source.connection
        return dict(connection=connection.connection_name,
                    name=project.name,
                    canonical_name=project.canonical_name,
                    override_branch=override_branch,
                    override_checkout=override_checkout,
                    default_branch=project_default_branch)

    if job.required_projects:
        for job_project in job.required_projects.values():
            (trusted, project) = tenant.getProject(
                job_project.project_name)
            if project is None:
                raise Exception("Unknown project %s" %
                                (job_project.project_name,))
            params['projects'].append(
                make_project_dict(project,
                                  job_project.override_branch,
                                  job_project.override_checkout))
            projects.add(project)
            required_projects.add(project)

    if job.include_vars:
        for iv in job.include_vars:
            source = connections.getSource(iv['connection'])
            project = source.getProject(iv['project'])
            if project not in projects:
                params['projects'].append(make_project_dict(project))
                projects.add(project)

    for dep_change in dependent_changes:
        try:
            (_, project) = item.manager.tenant.getProject(
                dep_change['project']['canonical_name'])
            if not project:
                raise KeyError()
        except Exception:
            # We have to find the project this way because it may not
            # be registered in the tenant (ie, a foreign project).
            source = connections.getSourceByCanonicalHostname(
                dep_change['project']['canonical_hostname'])
            project = source.getProject(dep_change['project']['name'])

        if not job.includesProject(project, change, item):
            continue
        if project not in projects:
            params['projects'].append(make_project_dict(project))
            projects.add(project)
    for p in projects:
        zuul_params['projects'][p.canonical_name] = (dict(
            name=p.name,
            short_name=p.name.split('/')[-1],
            # Duplicate this into the dict too, so that iterating
            # project.values() is easier for callers
            canonical_name=p.canonical_name,
            canonical_hostname=p.canonical_hostname,
            src_dir=make_src_dir(
                p.canonical_hostname,
                p.name,
                job.workspace_scheme),
            required=(p in required_projects),
        ))

    if item.event:
        params['zuul_event_id'] = item.event.zuul_event_id
    return params


def zuul_params_from_job(job):
    zuul_params = {
        "job": job.name,
        "voting": job.voting,
        "pre_timeout": job.pre_timeout,
        "timeout": job.timeout,
        "post_timeout": job.post_timeout,
        "jobtags": sorted(job.tags),
        "_inheritance_path": list(job.inheritance_path),
    }
    if job.artifact_data:
        zuul_params['artifacts'] = job.artifact_data
    if job.override_checkout:
        zuul_params['override_checkout'] = job.override_checkout
    return zuul_params



# Copyright 2012 Google Inc.
# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import requests
import threading
import time

TOKEN_URL = ('http://metadata.google.internal/computeMetadata/'
             'v1/instance/service-accounts/default/token')
REFRESH = 25
RETRY_INTERVAL = 5


class GCloudAuth(requests.auth.AuthBase):
    log = logging.getLogger('zuul.GerritConnection')

    def __init__(self, user, password):
        self.token = None
        self.expires = 0
        try:
            self.getToken()
        except Exception:
            self.log.exception("Error updating token:")
        self.update_thread = threading.Thread(target=self.update)
        self.update_thread.daemon = True
        self.update_thread.start()

    def update(self):
        while True:
            try:
                self._update()
            except Exception:
                self.log.exception("Error updating token:")
                time.sleep(5)

    def _update(self):
        now = time.time()
        expires = self.expires - REFRESH
        expires = max(expires, now + RETRY_INTERVAL)
        while now < expires:
            time.sleep(expires - now)
            now = time.time()
        self.getToken()

    def getToken(self):
        r = requests.get(TOKEN_URL, headers={'Metadata-Flavor': 'Google'})
        data = r.json()
        self.token = data['access_token']
        self.expires = time.time() + data['expires_in']

    def __call__(self, request):
        request.prepare_cookies({'o': self.token})
        return request



# Copyright 2017 Red Hat, Inc.
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import signal

import zuul.cmd
from zuul.lib.config import get_default
from zuul.lib import fingergw


class FingerGatewayApp(zuul.cmd.ZuulDaemonApp):
    '''
    Class for the daemon that will distribute any finger requests to the
    appropriate Zuul executor handling the specified build UUID.
    '''
    app_name = 'fingergw'
    app_description = 'The Zuul finger gateway.'

    def __init__(self):
        super(FingerGatewayApp, self).__init__()
        self.gateway = None

    def createParser(self):
        parser = super(FingerGatewayApp, self).createParser()
        self.addSubCommands(parser, fingergw.COMMANDS)
        return parser

    def exit_handler(self, signum, frame):
        if self.gateway:
            self.gateway.stop()

    def run(self):
        '''
        Main entry point for the FingerGatewayApp.

        Called by the main() method of the parent class.
        '''
        self.handleCommands()

        self.setup_logging('fingergw', 'log_config')
        self.log = logging.getLogger('zuul.fingergw')

        cmdsock = get_default(
            self.config, 'fingergw', 'command_socket',
            '/var/lib/zuul/%s.socket' % self.app_name)

        self.gateway = fingergw.FingerGateway(
            self.config,
            cmdsock,
            self.getPidFile(),
        )

        self.gateway.start()

        if self.args.nodaemon:
            signal.signal(signal.SIGTERM, self.exit_handler)

        while True:
            try:
                self.gateway.join()
                break
            except KeyboardInterrupt:
                print("Ctrl + C: asking gateway to exit nicely...\n")
                self.exit_handler(signal.SIGINT, None)


def main():
    FingerGatewayApp().main()



# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import concurrent.futures
import json
import logging
import time

import requests


class AzureAuth(requests.auth.AuthBase):
    AUTH_URL = "https://login.microsoftonline.com/{tenantId}/oauth2/token"
    # https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow#first-case-access-token-request-with-a-shared-secret
    # https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow#third-case-access-token-request-with-a-federated-credential

    def __init__(self, subscription_id, tenant_id, client_id,
                 client_secret, federated_token_file):
        self.log = logging.getLogger("azul.auth")
        self.subscription_id = subscription_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.tenant_id = tenant_id
        self.federated_token_file = federated_token_file
        self.token = None
        self.expiration = 0

    def refresh(self):
        if self.expiration - time.time() >= 60:
            return
        if self.federated_token_file:
            response = self.refresh_federated()
        else:
            response = self.refresh_secret()
        ret = response.json()
        if 'access_token' in ret:
            self.token = ret['access_token']
            self.expiration = float(ret['expires_on'])
        elif ('error_codes' in ret and 'error_description' in ret):
            raise AzureError(response.status_code,
                             ret['error_codes'][0],
                             ret['error_description'])
        else:
            raise Exception("Unknown error when authenticating to Azure")

    def load_from_token_file(self):
        with open(self.auth_location, encoding="utf-8") as f:
            credential = json.load(f)
        self.subscription_id = credential['subscriptionId']
        return credential

    def refresh_secret(self):
        self.log.debug('Refreshing secret authentication token')
        url = self.AUTH_URL.format(tenantId=self.tenant_id)
        data = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'resource': 'https://management.azure.com/',
        }
        return requests.post(url, data)

    def refresh_federated(self):
        self.log.debug('Refreshing federated authentication token')
        url = self.AUTH_URL.format(tenantId=self.tenant_id)
        with open(self.federated_token_file, encoding="utf-8") as f:
            token = f.read()
        data = {
            'grant_type': 'client_credentials',
            'client_id': self.client_id,
            'resource': 'https://management.azure.com/',
            'client_assertion_type':
            'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            'client_assertion': token,
        }
        return requests.post(url, data)

    def __call__(self, r):
        self.refresh()
        r.headers["authorization"] = "Bearer " + self.token
        return r


class AzureError(Exception):
    def __init__(self, status_code, error_code, message):
        super().__init__(message)
        self.error_code = error_code
        self.status_code = status_code


class AzureNotFoundError(AzureError):
    pass


class AzureCRUD:
    base_subscription_url = '/subscriptions/{subscriptionId}/'
    base_url = ''

    def __init__(self, cloud, **kw):
        self.cloud = cloud
        self.args = kw.copy()
        self.args.update(subscriptionId=self.cloud.subscription_id)

    def id(self, endpoint=None, **kw):
        if endpoint is None:
            endpoint = ''
        else:
            endpoint = '/' + endpoint
        url = (self.base_subscription_url + self.base_url + endpoint
               + '?api-version={apiVersion}')
        args = self.args.copy()
        args.update(kw)
        return url.format(**args)

    def url(self, endpoint=None, **kw):
        path = self.id(endpoint, **kw)
        return 'https://management.azure.com' + path

    def id_url(self, url, **kw):
        base_url = 'https://management.azure.com'
        url = base_url + url + '?api-version={apiVersion}'
        args = self.args.copy()
        args.update(kw)
        return url.format(**args)

    def get_by_id(self, resource_id):
        url = self.id_url(resource_id)
        return self.cloud.get(url)

    def _list(self, **kw):
        url = self.url(**kw)
        return self.cloud.paginate(self.cloud.get(url))

    def list(self):
        return self._list()

    def _get(self, **kw):
        url = self.url(**kw)
        return self.cloud.get(url)

    def _create(self, params, **kw):
        url = self.url(**kw)
        return self.cloud.put(url, params)

    def _delete(self, **kw):
        url = self.url(**kw)
        return self.cloud.delete(url)

    def _post(self, endpoint, params, **kw):
        url = self.url(endpoint=endpoint, **kw)
        return self.cloud.post(url, params)

    def delete_by_id(self, id):
        url = self.id_url(id)
        return self.cloud.delete(url)


class AzureResourceGroupsCRUD(AzureCRUD):
    base_url = 'resourcegroups/{resourceGroupName}'

    def list(self):
        return self._list(resourceGroupName='')

    def get(self, name):
        return self._get(resourceGroupName=name)

    def create(self, name, params):
        return self._create(params, resourceGroupName=name)

    def delete(self, name):
        return self._delete(resourceGroupName=name)


class AzureGroupedProviderCRUD(AzureCRUD):
    base_url = (
        'resourceGroups/{resourceGroupName}/providers/'
        '{providerId}/{resource}/{resourceName}')

    def list(self, resource_group_name):
        return self._list(resourceGroupName=resource_group_name,
                          resourceName='')

    def get(self, resource_group_name, name):
        return self._get(resourceGroupName=resource_group_name,
                         resourceName=name)

    def create(self, resource_group_name, name, params):
        return self._create(params,
                            resourceGroupName=resource_group_name,
                            resourceName=name)

    def delete(self, resource_group_name, name):
        return self._delete(resourceGroupName=resource_group_name,
                            resourceName=name)

    def post(self, resource_group_name, name, endpoint, params):
        return self._post(endpoint, params,
                          resourceGroupName=resource_group_name,
                          resourceName=name)


class AzureProviderCRUD(AzureCRUD):
    base_url = (
        'providers/{providerId}/{resource}/{resourceName}'
    )

    def list(self):
        return self._list(resourceName='')

    def get(self, name):
        return self._get(resourceName=name)

    def create(self, name, params):
        return self._create(params,
                            resourceName=name)

    def delete(self, name):
        return self._delete(resourceName=name)

    def post(self, name, endpoint, params):
        return self._post(endpoint, params,
                          resourceName=name)


class AzureNetworkCRUD(AzureCRUD):
    base_url = (
        'resourceGroups/{resourceGroupName}/providers/'
        'Microsoft.Network/virtualNetworks/{virtualNetworkName}/'
        '{resource}/{resourceName}')

    def list(self, resource_group_name, virtual_network_name):
        return self._list(resourceGroupName=resource_group_name,
                          virtualNetworkName=virtual_network_name,
                          resourceName='')

    def get(self, resource_group_name, virtual_network_name, name):
        return self._get(resourceGroupName=resource_group_name,
                         virtualNetworkName=virtual_network_name,
                         resourceName=name)

    def create(self, resource_group_name, virtual_network_name, name, params):
        return self._create(params,
                            resourceGroupName=resource_group_name,
                            virtualNetworkName=virtual_network_name,
                            resourceName=name)

    def delete(self, resource_group_name, virtual_network_name, name):
        return self._delete(resourceGroupName=resource_group_name,
                            virtualNetworkName=virtual_network_name,
                            resourceName=name)


class AzureLocationCRUD(AzureCRUD):
    base_url = (
        'providers/{providerId}/locations/{location}/{resource}')

    def list(self, location):
        return self._list(location=location)


class AzureDictResponse(dict):
    def __init__(self, response, *args):
        super().__init__(*args)
        self.response = response
        self.last_retry = time.time()


class AzureListResponse(list):
    def __init__(self, response, *args):
        super().__init__(*args)
        self.response = response
        self.last_retry = time.time()


class AzureCloud:
    TIMEOUT = 60

    def __init__(self, subscription_id=None, tenant_id=None,
                 client_id=None, client_secret=None,
                 federated_token_file=None):
        self.session = requests.Session()
        self.log = logging.getLogger("azul")
        self.auth = AzureAuth(subscription_id, tenant_id, client_id,
                              client_secret, federated_token_file)
        self.subscription_id = subscription_id or self.auth.subscription_id
        self.resource_group_network_interfaces = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.Network',
            resource='networkInterfaces',
            apiVersion='2020-11-01')
        self.network_interfaces = AzureProviderCRUD(
            self,
            providerId='Microsoft.Network',
            resource='networkInterfaces',
            apiVersion='2020-11-01')
        self.resource_group_public_ip_addresses = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.Network',
            resource='publicIPAddresses',
            apiVersion='2020-11-01')
        self.public_ip_addresses = AzureProviderCRUD(
            self,
            providerId='Microsoft.Network',
            resource='publicIPAddresses',
            apiVersion='2020-11-01')
        self.resource_group_virtual_machines = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='virtualMachines',
            apiVersion='2023-03-01')
        self.virtual_machines = AzureProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='virtualMachines',
            apiVersion='2023-03-01')
        self.resource_group_disks = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='disks',
            apiVersion='2020-06-30')
        self.disks = AzureProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='disks',
            apiVersion='2020-06-30')
        self.resource_group_images = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='images',
            apiVersion='2020-12-01')
        self.images = AzureProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='images',
            apiVersion='2020-12-01')
        self.resource_groups = AzureResourceGroupsCRUD(
            self,
            apiVersion='2020-06-01')
        self.subnets = AzureNetworkCRUD(
            self,
            resource='subnets',
            apiVersion='2020-07-01')
        self.compute_usages = AzureLocationCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='usages',
            apiVersion='2020-12-01')
        self.compute_skus = AzureProviderCRUD(
            self,
            providerId='Microsoft.Compute',
            resource='skus',
            apiVersion='2019-04-01')
        self.resource_group_managed_identities = AzureGroupedProviderCRUD(
            self,
            providerId='Microsoft.ManagedIdentity',
            resource='userAssignedIdentities',
            apiVersion='2023-01-31')
        self.managed_identities = AzureProviderCRUD(
            self,
            providerId='Microsoft.ManagedIdentity',
            resource='userAssignedIdentities',
            apiVersion='2023-01-31')

    def get(self, url, codes=[200]):
        return self.request('GET', url, None, codes)

    def put(self, url, data, codes=[200, 201, 202]):
        return self.request('PUT', url, data, codes)

    def post(self, url, data, codes=[200, 202]):
        return self.request('POST', url, data, codes)

    def delete(self, url, codes=[200, 201, 202, 204]):
        return self.request('DELETE', url, None, codes)

    def request(self, method, url, data, codes):
        self.log.debug('%s: %s %s' % (method, url, data))
        response = self.session.request(
            method, url, json=data,
            auth=self.auth, timeout=self.TIMEOUT,
            headers={'Accept': 'application/json',
                     'Accept-Encoding': 'gzip'})

        self.log.debug("Received headers: %s", response.headers)
        if response.status_code in codes:
            if len(response.text):
                self.log.debug("Received: %s", response.text)
                ret_data = response.json()
                if isinstance(ret_data, list):
                    return AzureListResponse(response, ret_data)
                else:
                    return AzureDictResponse(response, ret_data)
            self.log.debug("Empty response")
            return AzureDictResponse(response, {})
        err = response.json()
        self.log.error(response.text)
        if response.status_code == 404:
            raise AzureNotFoundError(
                response.status_code,
                err['error']['code'],
                err['error']['message'])
        else:
            raise AzureError(response.status_code,
                             err['error']['code'],
                             err['error']['message'])

    def paginate(self, data):
        ret = data['value']
        while 'nextLink' in data:
            data = self.get(data['nextLink'])
            ret += data['value']
        return ret

    def check_async_operation(self, response):
        resp = response.response
        location = resp.headers.get(
            'Azure-AsyncOperation',
            resp.headers.get('Location', None))
        if not location:
            self.log.debug("No async operation found")
            return None
        remain = (response.last_retry +
                  float(resp.headers.get('Retry-After', 2))) - time.time()
        self.log.debug("remain time %s", remain)
        if remain > 0:
            time.sleep(remain)
        response.last_retry = time.time()
        return self.get(location)

    def wait_for_async_operation(self, response, timeout=600):
        start = time.time()
        while True:
            if time.time() - start > timeout:
                raise Exception("Timeout waiting for async operation")
            ret = self.check_async_operation(response)
            if ret is None:
                return
            if ret['status'] == 'InProgress':
                continue
            if ret['status'] == 'Succeeded':
                return ret
            raise Exception("Unhandled async operation result: %s",
                            ret['status'])

    def upload_sas_chunk(self, url, start, end, data):
        if 'comp=page' not in url:
            url += '&comp=page'
        headers = {
            'x-ms-blob-type': 'PageBlob',
            'x-ms-page-write': 'Update',
            'Content-Length': str(len(data)),
            'Range': f'bytes={start}-{end}',
        }
        requests.put(url, headers=headers, data=data).raise_for_status()

    def _upload_chunk(self, url, start, end, data):
        attempts = 10
        for x in range(attempts):
            try:
                self.upload_sas_chunk(url, start, end, data)
                break
            except Exception:
                if x == attempts - 1:
                    raise
                else:
                    time.sleep(2 * x)

    def upload_page_blob_to_sas_url(self, url, file_object,
                                    pagesize=(4 * 1024 * 1024),
                                    concurrency=10):
        start = 0
        futures = set()
        if 'comp=page' not in url:
            url += '&comp=page'
        with concurrent.futures.ThreadPoolExecutor(
                max_workers=concurrency) as executor:
            while True:
                chunk = file_object.read(pagesize)
                if not chunk:
                    break
                end = start + len(chunk) - 1
                future = executor.submit(self._upload_chunk, url,
                                         start, end, chunk)
                start += len(chunk)
                futures.add(future)
                # Keep the pool of work supplied with data but without
                # reading the entire file into memory.
                if len(futures) >= (concurrency * 2):
                    (done, futures) = concurrent.futures.wait(
                        futures,
                        return_when=concurrent.futures.FIRST_COMPLETED)
            # We're done reading the file, wait for all uploads to finish
            (done, futures) = concurrent.futures.wait(futures)



"""Change patchset to string

Revision ID: cfc0dc45f341
Revises: ba4cdce9b18c
Create Date: 2018-01-09 16:44:31.506958

"""

# revision identifiers, used by Alembic.
revision = 'cfc0dc45f341'
down_revision = 'ba4cdce9b18c'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

BUILDSET_TABLE = 'zuul_buildset'

# Note: the following migration does not do what it was intended to do
# (change the column type to a string).  It was introduced with an
# error which instead caused the nullable flag to be set to true (even
# though it was already set), effectively making it a no-op.  To
# maintain compatibility with later versions of alembic, this has been
# updated to explicitly do now what it implicitly did then.  A later
# migration (19d3a3ebfe1d) corrects the error and changes the type.


def upgrade(table_prefix=''):
    op.alter_column(table_prefix + BUILDSET_TABLE,
                    'patchset',
                    nullable=True,
                    existing_nullable=True,
                    existing_type=sa.Integer)


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2020 Red Hat, inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import os
import subprocess
import tempfile
import textwrap

from kazoo.exceptions import NoNodeError

import zuul.model
from zuul.lib import yamlutil

from tests.base import (
    AnsibleZuulTestCase,
    iterate_timeout,
    simple_layout,
)
from tests.unit.test_web import BaseTestWeb
from tests.unit.test_launcher import LauncherBaseTestCase


def _split_pretty_table(output):
    lines = output.decode().split('\n')
    headers = [x.strip() for x in lines[1].split('|') if x != '']
    # Trim headers and last line of the table
    return [dict(zip(headers,
                     [x.strip() for x in l.split('|') if x != '']))
            for l in lines[3:-2]]


def _split_line_output(output):
    lines = output.decode().split('\n')
    info = {}
    for l in lines:
        if l.startswith('==='):
            continue
        try:
            key, value = l.split(':', 1)
            info[key] = value.strip()
        except ValueError:
            continue
    return info


class TestSmokeZuulClient(BaseTestWeb):
    def test_is_installed(self):
        """Test that the CLI is installed"""
        test_version = subprocess.check_output(
            ['zuul-client', '--version'],
            stderr=subprocess.STDOUT)
        self.assertTrue(b'Zuul-client version:' in test_version)


class TestZuulClientEncrypt(BaseTestWeb):
    """Test using zuul-client to encrypt secrets"""
    tenant_config_file = 'config/secrets/main.yaml'
    config_file = 'zuul-admin-web.conf'
    secret = {'password': 'zuul-client'}
    large_secret = {'key': (('a' * 79 + '\n') * 50)[:-1]}

    def setUp(self):
        super(TestZuulClientEncrypt, self).setUp()
        self.executor_server.hold_jobs_in_build = False

    def _getSecrets(self, job, pbtype):
        secrets = []
        build = self.getJobFromHistory(job)
        for pb in getattr(build.jobdir, pbtype):
            if pb.secrets_content:
                secrets.append(
                    yamlutil.ansible_unsafe_load(pb.secrets_content))
            else:
                secrets.append({})
        return secrets

    def _getClientCommand(self, *args):
        return ['zuul-client',
                '--zuul-url', self.base_url,
                *args]

    def test_encrypt_large_secret(self):
        """Test that we can use zuul-client to encrypt a large secret"""
        p = subprocess.Popen(
            self._getClientCommand(
                'encrypt',
                '--tenant', 'tenant-one', '--project', 'org/project2',
                '--secret-name', 'my_secret', '--field-name', 'key'),
            stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        p.stdin.write(
            str.encode(self.large_secret['key'])
        )
        output, error = p.communicate()
        p.stdin.close()
        self._test_encrypt(self.large_secret, output, error)

    def test_encrypt(self):
        """Test that we can use zuul-client to generate a project secret"""
        p = subprocess.Popen(
            self._getClientCommand(
                'encrypt',
                '--tenant', 'tenant-one', '--project', 'org/project2',
                '--secret-name', 'my_secret', '--field-name', 'password'),
            stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        p.stdin.write(
            str.encode(self.secret['password'])
        )
        output, error = p.communicate()
        p.stdin.close()
        self._test_encrypt(self.secret, output, error)

    def test_encrypt_outfile(self):
        """Test that we can use zuul-client to generate a project secret to a
        file"""
        outfile = tempfile.NamedTemporaryFile(delete=False)
        p = subprocess.Popen(
            self._getClientCommand(
                'encrypt',
                '--tenant', 'tenant-one', '--project', 'org/project2',
                '--secret-name', 'my_secret', '--field-name', 'password',
                '--outfile', outfile.name),
            stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        p.stdin.write(
            str.encode(self.secret['password'])
        )
        _, error = p.communicate()
        p.stdin.close()
        output = outfile.read()
        self._test_encrypt(self.secret, output, error)

    def test_encrypt_infile(self):
        """Test that we can use zuul-client to generate a project secret from
        a file"""
        infile = tempfile.NamedTemporaryFile(delete=False)
        infile.write(
            str.encode(self.secret['password'])
        )
        infile.close()
        p = subprocess.Popen(
            self._getClientCommand(
                'encrypt',
                '--tenant', 'tenant-one', '--project', 'org/project2',
                '--secret-name', 'my_secret', '--field-name', 'password',
                '--infile', infile.name),
            stdout=subprocess.PIPE)
        output, error = p.communicate()
        os.unlink(infile.name)
        self._test_encrypt(self.secret, output, error)

    def _test_encrypt(self, _secret, output, error):
        self.assertEqual(None, error, error)
        self.assertTrue(b'- secret:' in output, output.decode())
        new_repo_conf = output.decode()
        new_repo_conf += textwrap.dedent(
            """

            - job:
                parent: base
                name: project2-secret
                run: playbooks/secret.yaml
                secrets:
                  - my_secret

            - project:
                check:
                  jobs:
                    - project2-secret
                gate:
                  jobs:
                    - noop
            """
        )
        file_dict = {'zuul.yaml': new_repo_conf}
        A = self.fake_gerrit.addFakeChange('org/project2', 'master',
                                           'Add secret',
                                           files=file_dict)
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.assertEqual(A.data['status'], 'MERGED')
        self.fake_gerrit.addEvent(A.getChangeMergedEvent())
        self.waitUntilSettled()
        # check that the secret is used from there on
        B = self.fake_gerrit.addFakeChange('org/project2', 'master',
                                           'test secret',
                                           files={'newfile': 'xxx'})
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertEqual(B.reported, 1, "B should report success")
        self.assertHistory([
            dict(name='project2-secret', result='SUCCESS', changes='2,1'),
        ])
        secrets = self._getSecrets('project2-secret', 'playbooks')
        self.assertEqual(
            secrets,
            [{'my_secret': _secret}],
            secrets)


class TestZuulClientAdmin(BaseTestWeb):
    """Test the admin commands of zuul-client"""
    config_file = 'zuul-admin-web.conf'

    def test_autohold(self):
        """Test that autohold can be set with the Web client"""
        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'autohold', '--reason', 'some reason',
             '--tenant', 'tenant-one', '--project', 'org/project',
             '--job', 'project-test2', '--count', '1'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        # Check result
        resp = self.get_url(
            "api/tenant/tenant-one/autohold",
            headers={'Authorization': 'Bearer %s' % token})
        self.assertEqual(200, resp.status_code, resp.text)
        autohold_requests = resp.json()
        self.assertNotEqual([], autohold_requests)
        self.assertEqual(1, len(autohold_requests))
        request = autohold_requests[0]
        self.assertEqual('tenant-one', request['tenant'])
        self.assertIn('org/project', request['project'])
        self.assertEqual('project-test2', request['job'])
        self.assertEqual(".*", request['ref_filter'])
        self.assertEqual("some reason", request['reason'])
        self.assertEqual(1, request['max_count'])

    def test_enqueue(self):
        """Test that the Web client can enqueue a change"""
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.addApproval('Code-Review', 2)
        A.addApproval('Approved', 1)

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'enqueue', '--tenant', 'tenant-one',
             '--project', 'org/project',
             '--pipeline', 'gate', '--change', '1,1'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()
        # Check the build history for our enqueued build
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        # project-merge, project-test1, project-test2 in SUCCESS
        self.assertEqual(self.countJobResults(self.history, 'SUCCESS'), 3)

    def test_enqueue_ref(self):
        """Test that the Web client can enqueue a ref"""
        self.executor_server.hold_jobs_in_build = True
        p = "review.example.com/org/project"
        upstream = self.getUpstreamRepos([p])
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        A.setMerged()
        A_commit = str(upstream[p].commit('master'))
        self.log.debug("A commit: %s" % A_commit)

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'enqueue-ref', '--tenant', 'tenant-one',
             '--project', 'org/project',
             '--pipeline', 'post', '--ref', 'master',
             '--oldrev', '90f173846e3af9154517b88543ffbd1691f31366',
             '--newrev', A_commit],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()
        # Check the build history for our enqueued build
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertEqual(self.countJobResults(self.history, 'SUCCESS'), 1)

    def test_dequeue(self):
        """Test that the Web client can dequeue a change"""
        self.executor_server.hold_jobs_in_build = True
        start_builds = len(self.builds)
        self.create_branch('org/project', 'stable')
        self.fake_gerrit.addEvent(
            self.fake_gerrit.getFakeBranchCreatedEvent(
                'org/project', 'stable'))
        self.executor_server.hold_jobs_in_build = True
        self.commitConfigUpdate('common-config', 'layouts/timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()

        for _ in iterate_timeout(30, 'Wait for a build on hold'):
            if len(self.builds) > start_builds:
                break
        self.waitUntilSettled()

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'dequeue', '--tenant', 'tenant-one', '--project', 'org/project',
             '--pipeline', 'periodic', '--ref', 'refs/heads/stable'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()

        self.commitConfigUpdate('common-config',
                                'layouts/no-timer.yaml')
        self.scheds.execute(lambda app: app.sched.reconfigure(app.config))
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.assertEqual(self.countJobResults(self.history, 'ABORTED'), 1)

    def test_promote(self):
        "Test that the Web client can promote a change"
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        A.addApproval('Code-Review', 2)
        B.addApproval('Code-Review', 2)
        C.addApproval('Code-Review', 2)

        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.fake_gerrit.addEvent(C.addApproval('Approved', 1))

        self.waitUntilSettled()

        items = self.getAllItems('tenant-one', 'gate')
        enqueue_times = {}
        for item in items:
            enqueue_times[str(item.changes[0])] = item.enqueue_time

        # Promote B and C using the cli
        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'promote', '--tenant', 'tenant-one',
             '--pipeline', 'gate', '--changes', '2,1', '3,1'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()

        # ensure that enqueue times are durable
        items = self.getAllItems('tenant-one', 'gate')
        for item in items:
            self.assertEqual(
                enqueue_times[str(item.changes[0])], item.enqueue_time)

        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()
        self.executor_server.release('.*-merge')
        self.waitUntilSettled()

        self.assertEqual(len(self.builds), 6)
        self.assertEqual(self.builds[0].name, 'project-test1')
        self.assertEqual(self.builds[1].name, 'project-test2')
        self.assertEqual(self.builds[2].name, 'project-test1')
        self.assertEqual(self.builds[3].name, 'project-test2')
        self.assertEqual(self.builds[4].name, 'project-test1')
        self.assertEqual(self.builds[5].name, 'project-test2')

        self.assertTrue(self.builds[0].hasChanges(B))
        self.assertFalse(self.builds[0].hasChanges(A))
        self.assertFalse(self.builds[0].hasChanges(C))

        self.assertTrue(self.builds[2].hasChanges(B))
        self.assertTrue(self.builds[2].hasChanges(C))
        self.assertFalse(self.builds[2].hasChanges(A))

        self.assertTrue(self.builds[4].hasChanges(B))
        self.assertTrue(self.builds[4].hasChanges(C))
        self.assertTrue(self.builds[4].hasChanges(A))

        self.executor_server.release()
        self.waitUntilSettled()

        self.assertEqual(A.data['status'], 'MERGED')
        self.assertEqual(A.reported, 2)
        self.assertEqual(B.data['status'], 'MERGED')
        self.assertEqual(B.reported, 2)
        self.assertEqual(C.data['status'], 'MERGED')
        self.assertEqual(C.reported, 2)

    def test_manage_events(self):
        self.executor_server.hold_jobs_in_build = False
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
            dict(name='project-test1', result='SUCCESS', changes='1,1'),
            dict(name='project-test2', result='SUCCESS', changes='1,1'),
        ], ordered=False)

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'manage-events', '--tenant', 'tenant-one',
             'pause-trigger',
             '--reason', 'test trigger paused'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        time.sleep(1)

        B = self.fake_gerrit.addFakeChange('org/project', 'master', 'B')
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        time.sleep(5)

        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
            dict(name='project-test1', result='SUCCESS', changes='1,1'),
            dict(name='project-test2', result='SUCCESS', changes='1,1'),
        ], ordered=False)
        self.assertEqual(0, len(self.builds))

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'manage-events', '--tenant', 'tenant-one',
             'pause-result',
             '--reason', 'test result paused'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)

        time.sleep(5)
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
            dict(name='project-test1', result='SUCCESS', changes='1,1'),
            dict(name='project-test2', result='SUCCESS', changes='1,1'),
        ], ordered=False)
        self.assertEqual(0, len(self.builds))

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'manage-events', '--all-tenants', 'normal'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)

        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
            dict(name='project-test1', result='SUCCESS', changes='1,1'),
            dict(name='project-test2', result='SUCCESS', changes='1,1'),
            dict(name='project-merge', result='SUCCESS', changes='2,1'),
            dict(name='project-test1', result='SUCCESS', changes='2,1'),
            dict(name='project-test2', result='SUCCESS', changes='2,1'),
        ], ordered=False)

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token, '-v',
             'manage-events', '--all-tenants', 'discard-trigger'],
            stdout=subprocess.PIPE)
        output = p.communicate()
        self.assertEqual(p.returncode, 0, output)

        C = self.fake_gerrit.addFakeChange('org/project', 'master', 'C')
        self.fake_gerrit.addEvent(C.getPatchsetCreatedEvent(1))

        self.waitUntilSettled()
        self.assertHistory([
            dict(name='project-merge', result='SUCCESS', changes='1,1'),
            dict(name='project-test1', result='SUCCESS', changes='1,1'),
            dict(name='project-test2', result='SUCCESS', changes='1,1'),
            dict(name='project-merge', result='SUCCESS', changes='2,1'),
            dict(name='project-test1', result='SUCCESS', changes='2,1'),
            dict(name='project-test2', result='SUCCESS', changes='2,1'),
        ], ordered=False)


class TestZuulClientQueryData(BaseTestWeb):
    """Test that zuul-client can fetch builds"""
    config_file = 'zuul-sql-driver-mysql.conf'
    tenant_config_file = 'config/sql-driver/main.yaml'

    def setUp(self):
        super(TestZuulClientQueryData, self).setUp()
        self.add_base_changes()

    def add_base_changes(self):
        # change on org/project will run 5 jobs in check
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        B = self.fake_gerrit.addFakeChange('org/project1', 'master', 'B')
        # fail project-merge on PS1; its 2 dependent jobs will be skipped
        self.executor_server.failJob('project-merge', B)
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = True
        B.addPatchset()
        self.fake_gerrit.addEvent(B.getPatchsetCreatedEvent(2))
        # change on org/project1 will run 3 jobs in check
        self.waitUntilSettled()
        # changes on both projects will run 3 jobs in gate each
        A.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(A.addApproval('Approved', 1))
        B.addApproval('Code-Review', 2)
        self.fake_gerrit.addEvent(B.addApproval('Approved', 1))
        self.waitUntilSettled()
        self.executor_server.hold_jobs_in_build = False
        self.executor_server.release()
        self.waitUntilSettled()


class TestZuulClientBuilds(TestZuulClientQueryData,
                           AnsibleZuulTestCase):
    """Test that zuul-client can fetch builds"""
    def test_get_builds(self):
        """Test querying builds"""

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(17, len(results), results)

        # 5 jobs in check, 3 jobs in gate
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--project', 'org/project', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(8, len(results), results)
        self.assertTrue(all(x['Project'] == 'org/project' for x in results),
                        results)

        # project-test1 is run 3 times in check, 2 times in gate
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--job', 'project-test1', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(5, len(results), results)
        self.assertTrue(all(x['Job'] == 'project-test1' for x in results),
                        results)

        # 3 builds in check for 2,1; 3 in check + 3 in gate for 2,2
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--change', '2', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(9, len(results), results)
        self.assertTrue(all(x['Change or Ref'].startswith('2,')
                            for x in results),
                        results)

        # 1,3 does not exist
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--change', '1',
             '--ref', '3', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(0, len(results), results)

        for result in ['SUCCESS', 'FAILURE']:
            p = subprocess.Popen(
                ['zuul-client',
                 '--zuul-url', self.base_url,
                 'builds', '--tenant', 'tenant-one', '--result', result, ],
                stdout=subprocess.PIPE)
            job_count = self.countJobResults(self.history, result)
            # noop job not included, must be added
            if result == 'SUCCESS':
                job_count += 1
            output, err = p.communicate()
            self.assertEqual(p.returncode, 0, output)
            results = _split_pretty_table(output)
            self.assertEqual(job_count, len(results), results)
            if len(results) > 0:
                self.assertTrue(all(x['Result'] == result for x in results),
                                results)

        # 6 jobs in gate
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--pipeline', 'gate', ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(6, len(results), results)
        self.assertTrue(all(x['Pipeline'] == 'gate' for x in results),
                        results)


class TestZuulClientBuildInfo(TestZuulClientQueryData,
                              AnsibleZuulTestCase):
    """Test that zuul-client can fetch a build's details"""
    def test_get_build_info(self):
        """Test querying a specific build"""

        test_build = self.history[-1]

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'build-info', '--tenant', 'tenant-one',
             '--uuid', test_build.uuid],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, (output, err))
        info = _split_line_output(output)
        self.assertEqual(test_build.uuid, info.get('UUID'), test_build)
        self.assertEqual(test_build.result, info.get('Result'), test_build)
        self.assertEqual(test_build.name, info.get('Job'), test_build)

    def test_get_build_artifacts(self):
        """Test querying a specific build's artifacts"""
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'builds', '--tenant', 'tenant-one', '--job', 'project-test1',
             '--limit', '1'],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        uuid = results[0]['ID']
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'build-info', '--tenant', 'tenant-one',
             '--uuid', uuid,
             '--show-artifacts'],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, (output, err))
        artifacts = _split_pretty_table(output)
        self.assertTrue(
            any(x['name'] == 'tarball' and
                x['url'] == 'http://example.com/tarball'
                for x in artifacts),
            output)
        self.assertTrue(
            any(x['name'] == 'docs' and
                x['url'] == 'http://example.com/docs'
                for x in artifacts),
            output)


class TestZuulClientJobGraph(BaseTestWeb):

    def test_job_graph(self):
        """Test the job-graph command"""
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'job-graph',
             '--tenant', 'tenant-one',
             '--pipeline', 'check',
             '--project', 'org/project1',
             '--branch', 'master',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, (output, err))
        results = _split_pretty_table(output)
        expected = [
            {'Job': 'project-merge', 'Dependencies': ''},
            {'Job': 'project-test1', 'Dependencies': 'project-merge'},
            {'Job': 'project-test2', 'Dependencies': 'project-merge'},
            {'Job': 'project1-project2-integration',
             'Dependencies': 'project-merge'}
        ]
        self.assertEqual(results, expected)

    def test_job_graph_dot(self):
        """Test the job-graph command dot output"""
        p = subprocess.Popen(
            ['zuul-client',
             '--format', 'dot',
             '--zuul-url', self.base_url,
             'job-graph',
             '--tenant', 'tenant-one',
             '--pipeline', 'check',
             '--project', 'org/project1',
             '--branch', 'master',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, (output, err))
        expected = textwrap.dedent('''\
        digraph job_graph {
          rankdir=LR;
          node [shape=box];
          "project-merge";
          "project-merge" -> "project-test1" [dir=back];
          "project-merge" -> "project-test2" [dir=back];
          "project-merge" -> "project1-project2-integration" [dir=back];
        }
        ''').encode('utf8')
        self.assertEqual(output.strip(), expected.strip())


class TestZuulClientFreezeJob(BaseTestWeb):
    def test_freeze_job(self):
        """Test the freeze-job command"""
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'freeze-job',
             '--tenant', 'tenant-one',
             '--pipeline', 'check',
             '--project', 'org/project1',
             '--branch', 'master',
             '--job', 'project-test1',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, (output, err))
        output = output.decode('utf8')
        for s in [
                'Job: project-test1',
                'Branch: master',
                'Ansible Version:',
                'Workspace Scheme: golang',
                ('gerrit:common-config:playbooks/project-test1.yaml'
                 '@master [trusted]'),
        ]:
            self.assertIn(s, output)


class TestZuulClientAdminWithAccessRules(TestZuulClientAdmin):
    """Test the admin commands of zuul-client with access rules"""
    config_file = 'zuul-admin-web.conf'
    tenant_config_file = 'config/single-tenant/main-access-rules.yaml'
    default_token_groups = ['users']


class TestZuulClientEncryptWithAccessRules(TestZuulClientEncrypt):
    """Test using zuul-client to encrypt secrets"""
    tenant_config_file = 'config/secrets/main-access-rules.yaml'

    def _getClientCommand(self, *args):
        token = self._getToken()
        return ['zuul-client',
                '--zuul-url', self.base_url,
                '--auth-token', token, '-v',
                *args]


class TestZuulClientNodesets(LauncherBaseTestCase, BaseTestWeb):
    """Test that zuul-client can fetch nodeset info"""
    config_file = 'zuul-connections-nodepool.conf'

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_node_list(self):
        self.executor_server.hold_jobs_in_build = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'node-list', '--tenant', 'tenant-one',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(1, len(results), results)
        r = results[0]
        self.assertIsNotNone(r['UUID'])
        self.assertEqual("['debian-normal']", r['Labels'])
        self.assertEqual('ssh', r['Connection'])
        self.assertEqual(
            'review.example.com%2Forg%2Fcommon-config/aws-us-east-1-main',
            r['Provider'])
        self.assertEqual('in-use', r['State'])
        self.assertIsNotNone(r['State Time'])

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_nodeset_request_list(self):
        self.hold_nodeset_requests_in_queue = True
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url,
             'nodeset-request-list', '--tenant', 'tenant-one',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        results = _split_pretty_table(output)
        self.assertEqual(1, len(results), results)
        r = results[0]
        self.assertIsNotNone(r['UUID'])
        self.assertEqual("['debian-normal']", r['Labels'])
        self.assertIsNotNone(r['Buildset'])
        self.assertEqual('test-hold', r['State'])
        self.assertIsNotNone(r['Request Time'])
        self.assertEqual('check', r['Pipeline'])
        self.assertEqual('check-job', r['Job'])

        self.hold_nodeset_requests_in_queue = False
        self.executor_server.hold_jobs_in_build = False
        self.releaseNodesetRequests()
        self.waitUntilSettled()

    @simple_layout('layouts/nodepool.yaml', enable_nodepool=True)
    def test_nodes_hold_delete(self):
        # This tests 3 things (since they are lifecycle related):
        # * Setting the node state to hold
        # * Setting the node state to delete
        # * Deleting the request
        self.waitUntilSettled()

        request = self.requestNodes(['debian-normal'])
        self.assertEqual(request.state,
                         zuul.model.NodesetRequest.State.FULFILLED)
        self.assertEqual(len(request.nodes), 1)

        nodes = self.get_url('api/tenant/tenant-one/nodes').json()
        self.assertEqual(len(nodes), 1)

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token,
             'node-set-state', '--tenant', 'tenant-one',
             nodes[0]['uuid'], 'hold',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()

        node = self.launcher.api.nodes_cache.getItem(nodes[0]['uuid'])
        self.assertEqual(node.State.HOLD, node.state)

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token,
             'node-set-state', '--tenant', 'tenant-one',
             nodes[0]['uuid'], 'used',
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()

        requests = self.get_url(
            'api/tenant/tenant-one/nodeset-requests').json()
        self.assertEqual(len(requests), 1)
        self.assertEqual(request.uuid, requests[0]['uuid'])

        token = self._getToken()
        p = subprocess.Popen(
            ['zuul-client',
             '--zuul-url', self.base_url, '--auth-token', token,
             'nodeset-request-delete', '--tenant', 'tenant-one',
             requests[0]['uuid'],
             ],
            stdout=subprocess.PIPE)
        output, err = p.communicate()
        self.assertEqual(p.returncode, 0, output)
        self.waitUntilSettled()

        ctx = self.createZKContext(None)
        for _ in iterate_timeout(10, "request to be deleted"):
            try:
                request.refresh(ctx)
            except NoNodeError:
                break
        for _ in iterate_timeout(10, "node to be deleted"):
            try:
                node.refresh(ctx)
            except NoNodeError:
                break



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math
import psutil

from zuul.executor.sensors import SensorInterface
from zuul.lib.config import get_default

CGROUP_STATS_FILE = '/sys/fs/cgroup/memory/memory.stat'


def get_avail_mem_pct():
    avail_mem_pct = 100.0 - psutil.virtual_memory().percent
    return avail_mem_pct


class RAMSensor(SensorInterface):
    log = logging.getLogger("zuul.executor.sensor.ram")

    def __init__(self, statsd, base_key, config=None):
        super().__init__(statsd, base_key)
        self.min_avail_mem = float(get_default(config, 'executor',
                                               'min_avail_mem', '5.0'))
        self.cgroup_stats_file = CGROUP_STATS_FILE

    def _read_cgroup_stat(self):
        stat = {}
        try:
            with open(self.cgroup_stats_file) as f:
                for line in f.readlines():
                    key, value = line.split(' ')
                    stat[key] = int(value.strip())
        except Exception:
            pass
        return stat

    def _get_cgroup_limit(self):
        stat = self._read_cgroup_stat()
        limit = stat.get('hierarchical_memory_limit', math.inf)
        mem_total = psutil.virtual_memory().total
        if limit < mem_total:
            return limit
        else:
            return math.inf

    def _get_avail_mem_pct_cgroup(self):
        stat = self._read_cgroup_stat()
        limit = stat.get('hierarchical_memory_limit', math.inf)
        usage = stat.get('total_rss', math.inf)

        if math.isinf(limit) or math.isinf(usage):
            # pretend we have all memory available if we got infs
            return 100

        return 100.0 - usage / limit * 100

    def isOk(self):
        avail_mem_pct = get_avail_mem_pct()
        avail_mem_pct_cgroup = self._get_avail_mem_pct_cgroup()

        if self.statsd:
            self.statsd.gauge(self.base_key + '.pct_used_ram',
                              int((100.0 - avail_mem_pct) * 100))
            if math.isfinite(self._get_cgroup_limit()):
                self.statsd.gauge(self.base_key + '.pct_used_ram_cgroup',
                                  int((100.0 - avail_mem_pct_cgroup) * 100))

        if avail_mem_pct < self.min_avail_mem:
            return False, "low memory {:3.1f}% < {}".format(
                avail_mem_pct, self.min_avail_mem)

        if math.isinf(self._get_cgroup_limit()):
            # we have no cgroup defined limit so we're done now
            return True, "memory {:3.1f}% >= {}%".format(
                avail_mem_pct, self.min_avail_mem)

        if avail_mem_pct_cgroup < self.min_avail_mem:
            return False, "low memory cgroup {:3.1f}% < {}%".format(
                avail_mem_pct_cgroup, self.min_avail_mem)

        return True, "memory cgroup {:3.1f}% >= {}%, {:3.1f}% >= {}%".format(
            avail_mem_pct, self.min_avail_mem,
            avail_mem_pct_cgroup, self.min_avail_mem)
